"""Windows Imaging Component library definitions for Windows OS."""
from ctypes import byref, c_void_p, memmove

from pyglet.image import ImageData
from pyglet.image.codecs import ImageDecodeException, ImageDecoder, ImageEncoder, os
from pyglet.libs.win32 import _kernel32 as kernel32
from pyglet.libs.win32 import _ole32 as ole32
from pyglet.libs.win32 import com
from pyglet.libs.win32.constants import (
    CLSCTX_INPROC_SERVER,
    GENERIC_WRITE,
    GMEM_MOVEABLE,
    STREAM_SEEK_SET,
    WINDOWS_8_OR_GREATER,
)
from pyglet.libs.win32.types import (
    BOOL,
    BYTE,
    DOUBLE,
    DWORD,
    LPCWSTR,
    POINTER,
    PROPVARIANT,
    STATSTG,
    UINT,
    ULONG,
    IStream,
)

CLSID_WICImagingFactory1 = com.GUID(0xcacaf262, 0x9370, 0x4615, 0xa1, 0x3b, 0x9f, 0x55, 0x39, 0xda, 0x4c, 0xa)
CLSID_WICImagingFactory2 = com.GUID(0x317d06e8, 0x5f24, 0x433d, 0xbd, 0xf7, 0x79, 0xce, 0x68, 0xd8, 0xab, 0xc2)


WICBitmapCreateCacheOption = UINT
WICBitmapNoCache = 0
WICBitmapCacheOnDemand = 0x1
WICBitmapCacheOnLoad = 0x2
WICBITMAPCREATECACHEOPTION_FORCE_DWORD = 0x7fffffff

WICBitmapPaletteType = UINT
WICBitmapPaletteTypeCustom = 0

WICBitmapTransformOptions = UINT
WICBitmapTransformRotate0 = 0
WICBitmapTransformRotate90 = 0x1
WICBitmapTransformRotate180 = 0x2
WICBitmapTransformRotate270 = 0x3
WICBitmapTransformFlipHorizontal = 0x8
WICBitmapTransformFlipVertical = 0x10

WICBitmapDitherType = UINT
WICBitmapDitherTypeNone = 0
WICBitmapDitherTypeSolid = 0
WICBitmapDitherTypeOrdered4x4 = 0x1
WICBitmapDitherTypeOrdered8x8 = 0x2
WICBitmapDitherTypeOrdered16x16 = 0x3
WICBitmapDitherTypeSpiral4x4 = 0x4
WICBitmapDitherTypeSpiral8x8 = 0x5
WICBitmapDitherTypeDualSpiral4x4 = 0x6
WICBitmapDitherTypeDualSpiral8x8 = 0x7
WICBitmapDitherTypeErrorDiffusion = 0x8
WICBITMAPDITHERTYPE_FORCE_DWORD = 0x7fffffff
WICBITMAPTRANSFORMOPTIONS_FORCE_DWORD = 0x7fffffff

WICDecodeOptions = UINT
WICDecodeMetadataCacheOnDemand = 0
WICDecodeMetadataCacheOnLoad = 0x1
WICMETADATACACHEOPTION_FORCE_DWORD = 0x7fffffff

WICBitmapEncoderCacheOption = UINT
WICBitmapEncoderCacheInMemory = 0x0
WICBitmapEncoderCacheTempFile = 0x1
WICBitmapEncoderNoCache = 0x2
WICBITMAPENCODERCACHEOPTION_FORCE_DWORD = 0x7fffffff

# Different pixel formats.
REFWICPixelFormatGUID = POINTER(com.GUID)
GUID_WICPixelFormatDontCare = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x00)
GUID_WICPixelFormat1bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x01)
GUID_WICPixelFormat2bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x02)
GUID_WICPixelFormat4bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x03)
GUID_WICPixelFormat8bppIndexed = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x04)
GUID_WICPixelFormatBlackWhite = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x05)
GUID_WICPixelFormat2bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x06)
GUID_WICPixelFormat4bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x07)
GUID_WICPixelFormat8bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x08)
GUID_WICPixelFormat8bppAlpha = com.GUID(0xe6cd0116, 0xeeba, 0x4161, 0xaa, 0x85, 0x27, 0xdd, 0x9f, 0xb3, 0xa8, 0x95)
GUID_WICPixelFormat16bppBGR555 = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x09)
GUID_WICPixelFormat16bppBGR565 = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0a)
GUID_WICPixelFormat16bppBGRA5551 = com.GUID(0x05ec7c2b, 0xf1e6, 0x4961, 0xad, 0x46, 0xe1, 0xcc, 0x81, 0x0a, 0x87, 0xd2)
GUID_WICPixelFormat16bppGray = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0b)
GUID_WICPixelFormat24bppBGR = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0c)
GUID_WICPixelFormat24bppRGB = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0d)
GUID_WICPixelFormat32bppBGR = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0e)
GUID_WICPixelFormat32bppBGRA = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x0f)
GUID_WICPixelFormat32bppPBGRA = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x10)
GUID_WICPixelFormat32bppRGB = com.GUID(0xd98c6b95, 0x3efe, 0x47d6, 0xbb, 0x25, 0xeb, 0x17, 0x48, 0xab, 0x0c,
                                       0xf1)  # 7 platform update?
GUID_WICPixelFormat32bppRGBA = com.GUID(0xf5c7ad2d, 0x6a8d, 0x43dd, 0xa7, 0xa8, 0xa2, 0x99, 0x35, 0x26, 0x1a, 0xe9)
GUID_WICPixelFormat32bppPRGBA = com.GUID(0x3cc4a650, 0xa527, 0x4d37, 0xa9, 0x16, 0x31, 0x42, 0xc7, 0xeb, 0xed, 0xba)
GUID_WICPixelFormat48bppRGB = com.GUID(0x6fddc324, 0x4e03, 0x4bfe, 0xb1, 0x85, 0x3d, 0x77, 0x76, 0x8d, 0xc9, 0x15)
GUID_WICPixelFormat48bppBGR = com.GUID(0xe605a384, 0xb468, 0x46ce, 0xbb, 0x2e, 0x36, 0xf1, 0x80, 0xe6, 0x43, 0x13)

GUID_ContainerFormatBmp = com.GUID(0x0af1d87e, 0xfcfe, 0x4188, 0xbd, 0xeb, 0xa7, 0x90, 0x64, 0x71, 0xcb, 0xe3)
GUID_ContainerFormatPng = com.GUID(0x1b7cfaf4, 0x713f, 0x473c, 0xbb, 0xcd, 0x61, 0x37, 0x42, 0x5f, 0xae, 0xaf)
GUID_ContainerFormatIco = com.GUID(0xa3a860c4, 0x338f, 0x4c17, 0x91, 0x9a, 0xfb, 0xa4, 0xb5, 0x62, 0x8f, 0x21)
GUID_ContainerFormatJpeg = com.GUID(0x19e4a5aa, 0x5662, 0x4fc5, 0xa0, 0xc0, 0x17, 0x58, 0x02, 0x8e, 0x10, 0x57)
GUID_ContainerFormatTiff = com.GUID(0x163bcc30, 0xe2e9, 0x4f0b, 0x96, 0x1d, 0xa3, 0xe9, 0xfd, 0xb7, 0x88, 0xa3)
GUID_ContainerFormatGif = com.GUID(0x1f8a5601, 0x7d4d, 0x4cbd, 0x9c, 0x82, 0x1b, 0xc8, 0xd4, 0xee, 0xb9, 0xa5)
GUID_ContainerFormatWmp = com.GUID(0x57a37caa, 0x367a, 0x4540, 0x91, 0x6b, 0xf1, 0x83, 0xc5, 0x09, 0x3a, 0x4b)


class IPropertyBag2(com.pIUnknown):
    _methods_ = [
        ('Read',
         com.STDMETHOD()),
        ('Write',
         com.STDMETHOD()),
        ('CountProperties',
         com.STDMETHOD()),
        ('GetPropertyInfo',
         com.STDMETHOD()),
        ('LoadObject',
         com.STDMETHOD())
    ]


class IWICPalette(com.pIUnknown):
    _methods_ = [
        ('InitializePredefined',
         com.STDMETHOD()),
        ('InitializeCustom',
         com.STDMETHOD()),
        ('InitializeFromBitmap',
         com.STDMETHOD()),
        ('InitializeFromPalette',
         com.STDMETHOD()),
        ('GetType',
         com.STDMETHOD()),
        ('GetColorCount',
         com.STDMETHOD()),
        ('GetColors',
         com.STDMETHOD()),
        ('IsBlackWhite',
         com.STDMETHOD()),
        ('IsGrayscale',
         com.STDMETHOD()),
        ('HasAlpha',
         com.STDMETHOD()),
    ]


class IWICStream(IStream, com.pIUnknown):
    _methods_ = [
        ('InitializeFromIStream',
         com.STDMETHOD(IStream)),
        ('InitializeFromFilename',
         com.STDMETHOD(LPCWSTR, DWORD)),
        ('InitializeFromMemory',
         com.STDMETHOD(POINTER(BYTE), DWORD)),
        ('InitializeFromIStreamRegion',
         com.STDMETHOD()),
    ]


class IWICBitmapFrameEncode(com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IPropertyBag2)),
        ('SetSize',
         com.STDMETHOD(UINT, UINT)),
        ('SetResolution',
         com.STDMETHOD()),
        ('SetPixelFormat',
         com.STDMETHOD(REFWICPixelFormatGUID)),
        ('SetColorContexts',
         com.STDMETHOD()),
        ('SetPalette',
         com.STDMETHOD(IWICPalette)),
        ('SetThumbnail',
         com.STDMETHOD()),
        ('WritePixels',
         com.STDMETHOD(UINT, UINT, UINT, POINTER(BYTE))),
        ('WriteSource',
         com.STDMETHOD()),
        ('Commit',
         com.STDMETHOD()),
        ('GetMetadataQueryWriter',
         com.STDMETHOD())
    ]


class IWICBitmapEncoder(com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IWICStream, WICBitmapEncoderCacheOption)),
        ('GetContainerFormat',
         com.STDMETHOD()),
        ('GetEncoderInfo',
         com.STDMETHOD()),
        ('SetColorContexts',
         com.STDMETHOD()),
        ('SetPalette',
         com.STDMETHOD()),
        ('SetThumbnail',
         com.STDMETHOD()),
        ('SetPreview',
         com.STDMETHOD()),
        ('CreateNewFrame',
         com.STDMETHOD(POINTER(IWICBitmapFrameEncode), POINTER(IPropertyBag2))),
        ('Commit',
         com.STDMETHOD()),
        ('GetMetadataQueryWriter',
         com.STDMETHOD())
    ]


class IWICComponentInfo(com.pIUnknown):
    _methods_ = [
        ('GetComponentType',
         com.STDMETHOD()),
        ('GetCLSID',
         com.STDMETHOD()),
        ('GetSigningStatus',
         com.STDMETHOD()),
        ('GetAuthor',
         com.STDMETHOD()),
        ('GetVendorGUID',
         com.STDMETHOD()),
        ('GetVersion',
         com.STDMETHOD()),
        ('GetSpecVersion',
         com.STDMETHOD()),
        ('GetFriendlyName',
         com.STDMETHOD())
    ]


class IWICPixelFormatInfo(IWICComponentInfo, com.pIUnknown):
    _methods_ = [
        ('GetFormatGUID',
         com.STDMETHOD(POINTER(com.GUID))),
        ('GetColorContext',
         com.STDMETHOD()),
        ('GetBitsPerPixel',
         com.STDMETHOD(POINTER(UINT))),
        ('GetChannelCount',
         com.STDMETHOD(POINTER(UINT))),
        ('GetChannelMask',
         com.STDMETHOD())
    ]


class IWICBitmapSource(com.pIUnknown):
    _methods_ = [
        ('GetSize',
         com.STDMETHOD(POINTER(UINT), POINTER(UINT))),
        ('GetPixelFormat',
         com.STDMETHOD(REFWICPixelFormatGUID)),
        ('GetResolution',
         com.STDMETHOD(POINTER(DOUBLE), POINTER(DOUBLE))),
        ('CopyPalette',
         com.STDMETHOD()),
        ('CopyPixels',
         com.STDMETHOD(c_void_p, UINT, UINT, c_void_p)),
    ]


class IWICFormatConverter(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IWICBitmapSource, REFWICPixelFormatGUID, WICBitmapDitherType, c_void_p, DOUBLE,
                       WICBitmapPaletteType)),
        ('CanConvert',
         com.STDMETHOD(REFWICPixelFormatGUID, REFWICPixelFormatGUID, POINTER(BOOL))),
    ]


class IWICMetadataQueryReader(com.pIUnknown):
    _methods_ = [
        ('GetContainerFormat',
         com.STDMETHOD()),
        ('GetLocation',
         com.STDMETHOD()),
        ('GetMetadataByName',
         com.STDMETHOD(LPCWSTR, c_void_p)),
        ('GetEnumerator',
         com.STDMETHOD()),
    ]


class IWICBitmapFrameDecode(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('GetMetadataQueryReader',
         com.STDMETHOD(POINTER(IWICMetadataQueryReader))),
        ('GetColorContexts',
         com.STDMETHOD()),
        ('GetThumbnail',
         com.STDMETHOD(POINTER(IWICBitmapSource))),
    ]


class IWICBitmapFlipRotator(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Initialize',
         com.STDMETHOD(IWICBitmapSource, WICBitmapTransformOptions)),
    ]


class IWICBitmap(IWICBitmapSource, com.pIUnknown):
    _methods_ = [
        ('Lock',
         com.STDMETHOD()),
        ('SetPalette',
         com.STDMETHOD()),
        ('SetResolution',
         com.STDMETHOD())
    ]


class IWICBitmapDecoder(com.pIUnknown):
    _methods_ = [
        ('QueryCapability',
         com.STDMETHOD()),
        ('Initialize',
         com.STDMETHOD()),
        ('GetContainerFormat',
         com.STDMETHOD()),
        ('GetDecoderInfo',
         com.STDMETHOD()),
        ('CopyPalette',
         com.STDMETHOD()),
        ('GetMetadataQueryReader',
         com.STDMETHOD(POINTER(IWICMetadataQueryReader))),
        ('GetPreview',
         com.STDMETHOD()),
        ('GetColorContexts',
         com.STDMETHOD()),
        ('GetThumbnail',
         com.STDMETHOD()),
        ('GetFrameCount',
         com.STDMETHOD(POINTER(UINT))),
        ('GetFrame',
         com.STDMETHOD(UINT, POINTER(IWICBitmapFrameDecode))),
    ]


IID_IWICImagingFactory1 = com.GUID(0xec5ec8a9, 0xc395, 0x4314, 0x9c, 0x77, 0x54, 0xd7, 0xa9, 0x35, 0xff, 0x70)
IID_IWICImagingFactory2 = com.GUID(0x7B816B45, 0x1996, 0x4476, 0xB1, 0x32, 0xDE, 0x9E, 0x24, 0x7C, 0x8A, 0xF0)


IID_IWICPixelFormatInfo = com.GUID(0xE8EDA601, 0x3D48, 0x431a, 0xAB, 0x44, 0x69, 0x05, 0x9B, 0xE8, 0x8B, 0xBE)


class IWICImagingFactory(com.pIUnknown):
    _methods_ = [
        ('CreateDecoderFromFilename',
         com.STDMETHOD(LPCWSTR, com.GUID, DWORD, WICDecodeOptions, POINTER(IWICBitmapDecoder))),
        ('CreateDecoderFromStream',
         com.STDMETHOD(com.pIUnknown, c_void_p, WICDecodeOptions, POINTER(IWICBitmapDecoder))),
        ('CreateDecoderFromFileHandle',
         com.STDMETHOD()),
        ('CreateComponentInfo',
         com.STDMETHOD(com.GUID, POINTER(IWICComponentInfo))),
        ('CreateDecoder',
         com.STDMETHOD()),
        ('CreateEncoder',
         com.STDMETHOD(POINTER(com.GUID), POINTER(com.GUID), POINTER(IWICBitmapEncoder))),
        ('CreatePalette',
         com.STDMETHOD(POINTER(IWICPalette))),
        ('CreateFormatConverter',
         com.STDMETHOD(POINTER(IWICFormatConverter))),
        ('CreateBitmapScaler',
         com.STDMETHOD()),
        ('CreateBitmapClipper',
         com.STDMETHOD()),
        ('CreateBitmapFlipRotator',
         com.STDMETHOD(POINTER(IWICBitmapFlipRotator))),
        ('CreateStream',
         com.STDMETHOD(POINTER(IWICStream))),
        ('CreateColorContext',
         com.STDMETHOD()),
        ('CreateColorTransformer',
         com.STDMETHOD()),
        ('CreateBitmap',
         com.STDMETHOD(UINT, UINT, REFWICPixelFormatGUID, WICBitmapCreateCacheOption, POINTER(IWICBitmap))),
        ('CreateBitmapFromSource',
         com.STDMETHOD()),
        ('CreateBitmapFromSourceRect',
         com.STDMETHOD()),
        ('CreateBitmapFromMemory',
         com.STDMETHOD(UINT, UINT, REFWICPixelFormatGUID, UINT, UINT, POINTER(BYTE), POINTER(IWICBitmap))),
        ('CreateBitmapFromHBITMAP',
         com.STDMETHOD()),
        ('CreateBitmapFromHICON',
         com.STDMETHOD()),
        ('CreateComponentEnumerator',
         com.STDMETHOD()),
        ('CreateFastMetadataEncoderFromDecoder',
         com.STDMETHOD()),
        ('CreateFastMetadataEncoderFromFrameDecode',
         com.STDMETHOD()),
        ('CreateQueryWriter',
         com.STDMETHOD()),
        ('CreateQueryWriterFromReader',
         com.STDMETHOD())
    ]

class IWICImagingFactory2(IWICImagingFactory):
    _methods_ = [
        ('CreateImageEncoder',
         com.STDMETHOD()),
    ]



# Experimental GPU based particle system

from __future__ import annotations

import sys
import time

import pyglet
from pyglet import clock, event, graphics, image
from pyglet.gl import *

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run

vertex_source = """#version 150
    in vec3 position;
    in vec4 size;
    in vec4 scale;
    in vec4 velocity;
    in vec4 color_start;
    in vec4 color_end;
    in vec4 texture_uv;
    in float rotation;
    in float birth;

    out vec4 geo_size;
    out vec4 geo_scale;
    out vec4 geo_velocity;
    out vec4 geo_color_start;
    out vec4 geo_color_end;
    out vec4 geo_tex_coords;
    out float geo_rotation;
    out float geo_birth;
    out int geo_vert_id;

    void main() {
        gl_Position = vec4(position, 1);
        geo_size = size;
        geo_scale = scale;
        geo_velocity = velocity;
        geo_color_start = color_start;
        geo_color_end = color_end;
        geo_tex_coords = texture_uv;
        geo_rotation = rotation;
        geo_birth = birth;
        geo_vert_id = gl_VertexID;
    }
"""

geometry_source = """#version 150
    // We are taking single points from the vertex shader
    // and emitting 4 new vertices to create a quad.
    layout (points) in;
    layout (triangle_strip, max_vertices = 32) out;

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;

    uniform float time;

    // Since geometry shader can take multiple values from a vertex
    // shader we need to define the inputs from it as arrays.
    // For our purposes, we just take single values (points).
    in vec4 geo_size[];
    in vec4 geo_scale[];
    in vec4 geo_velocity[];
    in vec4 geo_color_start[];
    in vec4 geo_color_end[];
    in vec4 geo_tex_coords[];
    in float geo_rotation[];
    in float geo_birth[];
    in int geo_vert_id[];

    out vec2 uv;
    out vec4 frag_color;

    void main() {
        // Unpack the image size and anchor
        vec2 size = geo_size[0].xy;
        vec2 anchor = geo_size[0].zw;
        vec2 scale_start = geo_scale[0].xy;
        vec2 scale_end = geo_scale[0].zw;
        
        vec2 velocity = geo_velocity[0].xy;
        vec2 spread = geo_velocity[0].zw;

        float birth = geo_birth[0];
        float elapsed = time - birth;
        float repeater = mod(elapsed, 1.0);

        int vert_id = geo_vert_id[0];

        for(int i=0;i<8;++i){
            // TODO: user supplied rotation speed
            float time_scale = mod(elapsed - (i / 7.0), 1.0);
            float rotation = geo_rotation[0] + time_scale * 100;

            // TODO: user supplied X, Y velocities
            vec3 center = gl_in[0].gl_Position.xyz;
            center.x += time_scale * velocity.x * (spread.x * cos(vert_id + 1) * sin(i + 1));
            center.y += time_scale * velocity.y * (spread.y * sin(vert_id + 1) * cos(i + 1));
            
            // Interpolate between the start and end colors, based on the lifetime 
            // (end - start) * step + start
            frag_color = (geo_color_end[0] - geo_color_start[0]) * time_scale + geo_color_start[0]; 
    
            // Interpolate between the start and end scale, based on the lifetime 
            // (end - start) * step + start
            mat4 m_scale = mat4(1.0);
            m_scale[0][0] = ((scale_end - scale_start) * time_scale + scale_start).x;
            m_scale[1][1] = ((scale_end - scale_start) * time_scale + scale_start).y;

            // This matrix controls the actual position of the particles:
            mat4 m_translate = mat4(1.0);
            m_translate[3][0] = center.x;
            m_translate[3][1] = center.y;
            m_translate[3][2] = center.z;

            mat4 m_rotation = mat4(1.0);
            m_rotation[0][0] =  cos(radians(-rotation)); 
            m_rotation[0][1] =  sin(radians(-rotation));
            m_rotation[1][0] = -sin(radians(-rotation));
            m_rotation[1][1] =  cos(radians(-rotation));    
    
            // Final UV coords (left, bottom, right, top):
            float uv_l = geo_tex_coords[0].s;
            float uv_b = geo_tex_coords[0].t;
            float uv_r = geo_tex_coords[0].p;
            float uv_t = geo_tex_coords[0].q;
    
            // Emit a triangle strip to create a quad (4 vertices).
            // Prepare and reuse the transformation matrix and fragment color:
            mat4 m_pv = window.projection * window.view * m_translate * m_rotation * m_scale;

            // Upper left
            gl_Position = m_pv * vec4(vec2(0.0, size.y) - anchor, 0.0, 1.0);
            uv = vec2(uv_l, uv_t);
            EmitVertex();
    
            // lower left
            gl_Position = m_pv * vec4(vec2(0.0, 0.0) - anchor, 0.0, 1.0);
            uv = vec2(uv_l, uv_b);
            EmitVertex();
    
            // upper right
            gl_Position = m_pv * vec4(vec2(size.x, size.y) - anchor, 0.0, 1.0);
            uv = vec2(uv_r, uv_t);
            EmitVertex();
    
            // lower right
            gl_Position = m_pv * vec4(vec2(size.x, 0.0) - anchor, 0.0, 1.0);
            uv = vec2(uv_r, uv_b);
            EmitVertex();

            // We are done with this triangle strip now
            EndPrimitive();    

        }
    }
"""

fragment_source = """#version 150
    in vec2 uv;
    in vec4 frag_color;
    out vec4 final_color;

    uniform sampler2D particle_texture;

    void main() {
        final_color = texture(particle_texture, uv) * frag_color;
    }

"""


def get_default_shader():
    return pyglet.gl.current_context.create_program((vertex_source, 'vertex'),
                                                    (geometry_source, 'geometry'),
                                                    (fragment_source, 'fragment'))


class EmitterGroup(graphics.Group):
    def __init__(self, texture, blend_src, blend_dest, program, parent=None):
        super().__init__(parent=parent)
        self.texture = texture
        self.blend_src = blend_src
        self.blend_dest = blend_dest
        self.program = program

    def set_state(self):
        self.program.use()

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.texture.target, self.texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(self.blend_src, self.blend_dest)

    def unset_state(self):
        glDisable(GL_BLEND)
        self.program.stop()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.texture})"

    def __eq__(self, other):
        return (other.__class__ is self.__class__ and
                self.program is other.program and
                self.parent == other.parent and
                self.texture.target == other.texture.target and
                self.texture.id == other.texture.id and
                self.blend_src == other.blend_src and
                self.blend_dest == other.blend_dest)

    def __hash__(self):
        return hash((self.program, self.parent,
                     self.texture.id, self.texture.target,
                     self.blend_src, self.blend_dest))


class Emitter(event.EventDispatcher):
    _batch = None
    _animation = None
    _frame_index = 0
    _paused = False
    _rotation = 0
    _visible = True
    _vertex_list = None
    group_class = EmitterGroup

    def __init__(self, img, x, y, z, count, velocity, spread,
                 color_start=(255, 255, 255, 255), color_end=(255, 255, 255, 255),
                 scale_start=(1.0, 1.0), scale_end=(1.0, 1.0),
                 blend_src=GL_SRC_ALPHA, blend_dest=GL_ONE_MINUS_SRC_ALPHA,
                 batch=None, group=None, program=None):

        self._img = img
        self._x = x
        self._y = y
        self._z = z
        self._count = count
        self._velocity = velocity + spread

        self._color_start = color_start
        self._color_end = color_end
        self._scale_start = scale_start
        self._scale_end = scale_end

        if isinstance(img, image.Animation):
            self._animation = img
            self._texture = img.frames[0].image.get_texture()
            self._next_dt = img.frames[0].duration
            if self._next_dt:
                clock.schedule_once(self._animate, self._next_dt)
        else:
            self._texture = img.get_texture()

        self._program = program or get_default_shader()
        self._batch = batch or graphics.get_default_batch()
        self._user_group = group
        self._group = self.group_class(self._texture, blend_src, blend_dest, self.program, group)
        self._create_vertex_list()

    def _create_vertex_list(self):
        texture = self._texture
        count = self._count
        self._vertex_list = self.program.vertex_list(
            count, GL_POINTS, self._batch, self._group,
            position=('f', (self._x, self._y, self._z) * count),

            size=('f', (texture.width, texture.height, texture.anchor_x, texture.anchor_y) * count),
            scale=('f', (self._scale_start + self._scale_end) * count),

            velocity=('f', self._velocity * count),

            color_start=('Bn', self._color_start * count),
            color_end=('Bn', self._color_end * count),

            texture_uv=('f', texture.uv * count),
            rotation=('f', (self._rotation,) * count),
            birth=('f', (time.perf_counter(),) * count))

    @property
    def program(self):
        return self._program

    @program.setter
    def program(self, program):
        if self._program == program:
            return
        self._group = self.group_class(self._texture,
                                       self._group.blend_src,
                                       self._group.blend_dest,
                                       program,
                                       self._user_group)
        if (self._batch and
                self._batch.update_shader(self._vertex_list, GL_POINTS, self._group, program)):
            # Exit early if changing domain is not needed.
            return

        # Recreate vertex list.
        self._vertex_list.delete()
        self._create_vertex_list()

    def delete(self):
        """Force immediate removal of the emitter from video memory.

        This is often necessary when using batches, as the Python garbage
        collector will not necessarily call the finalizer immediately.
        """
        if self._animation:
            clock.unschedule(self._animate)
        self._vertex_list.delete()
        self._vertex_list = None
        self._texture = None
        self._group = None

    def _animate(self, dt):
        self._frame_index += 1
        if self._frame_index >= len(self._animation.frames):
            self._frame_index = 0
            self.dispatch_event('on_animation_end')
            if self._vertex_list is None:
                return  # Deleted in event handler.

        frame = self._animation.frames[self._frame_index]
        self._set_texture(frame.image.get_texture())

        if frame.duration is not None:
            duration = frame.duration - (self._next_dt - dt)
            duration = min(max(0.0, duration), frame.duration)
            clock.schedule_once(self._animate, duration)
            self._next_dt = duration
        else:
            self.dispatch_event('on_animation_end')

    def _set_texture(self, texture):
        if texture.id is not self._texture.id:
            self._group = self._group.__class__(texture,
                                                self._group.blend_src,
                                                self._group.blend_dest,
                                                self._group.program,
                                                self._group.parent)
            self._vertex_list.delete()
            self._texture = texture
            self._create_vertex_list()
        else:
            self._vertex_list.texture_uv[:] = texture.uv
        self._texture = texture

    @property
    def position(self) -> tuple[int | float, int | float, int | float]:
        return self._x, self._y, self._z

    @position.setter
    def position(self, position: tuple[int | float, int | float, int | float]):
        self._x, self._y, self._z = position
        self._vertex_list.position[:] = position

    if _is_pyglet_doc_run:
        def on_animation_end(self):
            """The emitter animation reached the final frame.

            The event is triggered only if the emitter has an animation, not an
            image. For looping animations, the event is triggered each time
            the animation loops.

            :event:
            """


Emitter.register_event_type('on_animation_end')


class ParticleManager:

    def __init__(self, img, lifespan, count, velocity,
                 spread=(10.0, 10.0),
                 color_start=(255, 255, 255, 255), color_end=(255, 255, 255, 255),
                 scale_start=(1.0, 1.0), scale_end=(1.0, 1.0),
                 batch=None, group=None):

        self._img = img
        self._lifespan = lifespan
        self._count = count
        self._velocity = velocity
        self._spread = spread
        self._color_start = color_start
        self._color_end = color_end
        self._scale_start = scale_start
        self._scale_end = scale_end

        self._batch = batch
        self._group = group
        self._program = get_default_shader()
        clock.schedule_interval(self._update_shader_time, 1 / 60)

        # TODO: remove debug
        self.total_number = 0
        self.total_label = pyglet.text.Label("particles: 0", 10, 10, dpi=256, color=(10, 200, 10), batch=batch)

    def _update_shader_time(self, dt):
        self._program['time'] = time.perf_counter()

    def _delete_callback(self, dt, emitter):
        emitter.delete()

        # TODO: remove debug
        self.total_number -= 1
        self.total_label.text = f"particles: {self.total_number * self._count * 8!s}"

    def create_emitter(self, x, y, z=0):
        emitter = Emitter(self._img, x, y, z, self._count, self._velocity, self._spread,
                          color_start=self._color_start, color_end=self._color_end,
                          scale_start=self._scale_start, scale_end=self._scale_end,
                          batch=self._batch, group=self._group)
        pyglet.clock.schedule_once(self._delete_callback, self._lifespan, emitter)

        # TODO: remove debug
        self.total_number += 1
        self.total_label.text = f"particles: {self.total_number * self._count * 8!s}"

        return emitter



import pyglet

window = pyglet.window.Window()
label = pyglet.text.Label('Hello, world',
                          font_name='Times New Roman',
                          font_size=36,
                          x=window.width//2, y=window.height//2,
                          anchor_x='center', anchor_y='center')


@window.event
def on_draw():
    window.clear()
    label.draw()


pyglet.app.run()



"""
Usage

    bokeh_timeline.py sample

Renders media player's internal state graphically using bokeh

Arguments
    sample: sample to report

The output will be written to session's output dir under
    reports/sample.timeline.html

Notice the plot can be zoomed live with the mouse wheel, but you must click
the button that looks as a distorted 'OP'; it also does pan with mouse drag.

Example

    bokeh_timeline.py small.mp4

will write the output to report/small.mp4.timeline.html
"""
# just in case, this version was developed and tested with bokeh vs 0.12.6 (win)

import os
import sys

from bokeh.layouts import column
from bokeh.models import LinearAxis, Range1d
from bokeh.plotting import figure, output_file, show

import fs
import mpexceptions
import timeline


def main(sample):
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    bokeh_render_timeline(pathserv, sample)


def bokeh_render_timeline(pathserv, sample):
    infile = pathserv.report_filename(sample, "timeline.pkl", False)
    if not os.path.isfile(infile):
        timeline.save_timeline(pathserv, sample, ".pkl")

    timeline_postprocessed = fs.pickle_load(infile)
    info = unpack_timeline(timeline_postprocessed)
    outfile = pathserv.report_filename(sample, "timeline.html", False)
    make_plot(info, outfile)


def unpack_timeline(timeline_postprocessed):
    timeline, current_time_nones, audio_time_nones = timeline_postprocessed
    (wall_times, pyglet_times, audio_times, current_times,
     frame_nums, rescheds) = zip(*timeline)

    if current_time_nones:
        x_vnones, y_vnones = zip(*current_time_nones)
    else:
        x_vnones, y_vnones = [], []

    if audio_time_nones:
        x_anones, y_anones = zip(*audio_time_nones)
    else:
        x_anones, y_anones = [], []

    return (wall_times, pyglet_times, audio_times,
            current_times, frame_nums, rescheds,
            x_vnones, y_vnones,
            x_anones, y_anones)


def make_plot(info, outfile):
    # prepare some data
    (wall_times, pyglet_times, audio_times,
     current_times, frame_nums, rescheds,
     x_vnones, y_vnones,
     x_anones, y_anones) = info

    # output to static HTML file
    output_file(outfile)

    # main plot
    p = figure(
       tools="pan,wheel_zoom,reset,save",
       y_axis_type="linear", y_range=[0.000, wall_times[-1]], title="timeline",
       x_axis_label='wall_time', y_axis_label='time',
       plot_width=600, plot_height=600
    )

    # add some renderers
    p.line(wall_times, wall_times, legend="wall_time")
    #p.line(wall_times, pyglet_times, legend="pyglet_time", line_width=3)
    p.line(wall_times, current_times, legend="current_times", line_color="red")
    p.line(wall_times, audio_times, legend="audio_times", line_color="orange", line_dash="4 4")

    p.circle(x_vnones, y_vnones, legend="current time nones", fill_color="green", size=8)
    p.circle(x_anones, y_anones, legend="audio time nones", fill_color="red", size=6)

    # secondary y-axis for frame_num
    p.extra_y_ranges = {"frame_num": Range1d(start=0, end=frame_nums[-1])}
    p.line(wall_times, frame_nums, legend="frame_num",
           line_color="black", y_range_name="frame_num")
    p.add_layout(LinearAxis(y_range_name="frame_num", axis_label="frame num"), 'left')

    p.legend.location = "bottom_right"
    # show the results
    #show(p)

    # secondary plot for rescheduling times
    q = figure(
       tools="pan,wheel_zoom,reset,save",
       y_axis_type="linear", y_range=[-0.3, 0.3], title="rescheduling time",
       x_axis_label='wall_time', y_axis_label='rescheduling time',
       plot_width=600, plot_height=150
    )
    q.line(wall_times, rescheds)

    show(column(p, q))


def usage():
    print(__doc__)
    sys.exit(1)


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) < 2:
        usage()
    sample = sys.argv[1]
    return sample


if __name__ == '__main__':
    sample = sysargs_to_mainargs()
    main(sample)




class ExceptionPygletMediaPlayerTesting(Exception):
    pass

class ExceptionUndefinedSamplesDir(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionSamplesDirDoesNotExist(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionSessionExistWithSameName(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionNoSessionIsActive(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionNoSessionWithThatName(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionAttemptToBreakRawDataProtection(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionAttemptToBrekReportsProtection(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionPlaylistFileDoNotExists(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionBadSampleInPlaylist(ExceptionPygletMediaPlayerTesting):
    def __init__(self, rejected):
        self.rejected = rejected
        super().__init__()

class ExceptionUnknownReport(ExceptionPygletMediaPlayerTesting):
    def __init__(self, rpt_name):
        self.rpt_name = rpt_name

class ExceptionNoDbgFilesPresent(ExceptionPygletMediaPlayerTesting):
    pass

class ExceptionUnknownOutputFormat(ExceptionPygletMediaPlayerTesting):
    def __init__(self, output_format):
        self.output_format = output_format



from __future__ import annotations

from typing import TYPE_CHECKING

from pyglet.util import CodecRegistry, Decoder, Encoder, DecodeException, EncodeException

if TYPE_CHECKING:
    from typing import BinaryIO
    from pyglet.model import Model, Scene


registry = CodecRegistry()
add_decoders = registry.add_decoders
add_encoders = registry.add_encoders
get_decoders = registry.get_decoders
get_encoders = registry.get_encoders


class ModelDecodeException(DecodeException):
    pass


class ModelEncodeException(EncodeException):
    pass


class ModelDecoder(Decoder):
    def decode(self, filename: str, file: BinaryIO | None) -> Scene:
        """Decode the given file object and return a Scene instance.

        Throws ModelDecodeException if there is an error. If passing
        a ``file`` object, ``filename`` should be provided as a hint
        to the file type.
        """
        raise NotImplementedError()


class ModelEncoder(Encoder):

    def encode(self, model: Model, filename: str, file: BinaryIO | None) -> None:
        """Encode the given model to the given file."""
        raise NotImplementedError()


def add_default_codecs() -> None:
    # Add all bundled codecs.

    try:
        from pyglet.model.codecs import obj
        registry.add_decoders(obj)
    except ImportError:
        pass

    try:
        from pyglet.model.codecs import gltf
        registry.add_decoders(gltf)
    except ImportError:
        pass



import pyglet
import math
from pyglet.window import key
from . import bullet, physicalobject, resources


class Player(physicalobject.PhysicalObject):
    """Physical object that responds to user input"""

    def __init__(self, *args, **kwargs):
        super().__init__(img=resources.player_image, *args, **kwargs)

        # Create a child sprite to show when the ship is thrusting
        self.engine_sprite = pyglet.sprite.Sprite(img=resources.engine_image, *args, **kwargs)
        self.engine_sprite.visible = False

        # Set some easy-to-tweak constants
        self.thrust = 300.0
        self.rotate_speed = 200.0
        self.bullet_speed = 700.0

        # Player should not collide with own bullets
        self.reacts_to_bullets = False

        # Tell the game handler about any event handlers
        self.key_handler = key.KeyStateHandler()
        self.event_handlers = [self, self.key_handler]

    def update(self, dt):
        # Do all the normal physics stuff
        super().update(dt)

        if self.key_handler[key.LEFT]:
            self.rotation -= self.rotate_speed * dt
        if self.key_handler[key.RIGHT]:
            self.rotation += self.rotate_speed * dt

        if self.key_handler[key.UP]:
            # Note: pyglet's rotation attributes are in "negative degrees"
            angle_radians = -math.radians(self.rotation)
            force_x = math.cos(angle_radians) * self.thrust * dt
            force_y = math.sin(angle_radians) * self.thrust * dt
            self.velocity_x += force_x
            self.velocity_y += force_y

            # If thrusting, update the engine sprite
            self.engine_sprite.rotation = self.rotation
            self.engine_sprite.x = self.x
            self.engine_sprite.y = self.y
            self.engine_sprite.visible = True
        else:
            # Otherwise, hide it
            self.engine_sprite.visible = False

    def on_key_press(self, symbol, modifiers):
        if symbol == key.SPACE:
            self.fire()

    def fire(self):
        # Note: pyglet's rotation attributes are in "negative degrees"
        angle_radians = -math.radians(self.rotation)

        # Create a new bullet just in front of the player
        ship_radius = self.image.width / 2
        bullet_x = self.x + math.cos(angle_radians) * ship_radius
        bullet_y = self.y + math.sin(angle_radians) * ship_radius
        new_bullet = bullet.Bullet(bullet_x, bullet_y, batch=self.batch)

        # Give it some speed
        bullet_vx = self.velocity_x + math.cos(angle_radians) * self.bullet_speed
        bullet_vy = self.velocity_y + math.sin(angle_radians) * self.bullet_speed
        new_bullet.velocity_x, new_bullet.velocity_y = bullet_vx, bullet_vy

        # Add it to the list of objects to be added to the game_objects list
        self.new_objects.append(new_bullet)

        # Play the bullet sound
        resources.bullet_sound.play()

    def delete(self):
        # We have a child sprite which must be deleted when this object
        # is deleted from batches, etc.
        self.engine_sprite.delete()
        super().delete()



from pyglet.gl import *


def draw_client_border(window):
    glClearColor(0, 0, 0, 1)
    glClear(GL_COLOR_BUFFER_BIT)

    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    glOrtho(0, window.width, 0, window.height, -1, 1)

    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()

    def rect(x1, y1, x2, y2):
        glBegin(GL_LINE_LOOP)
        glVertex2f(x1, y1)
        glVertex2f(x2, y1)
        glVertex2f(x2, y2)
        glVertex2f(x1, y2)
        glEnd()

    glColor3f(1, 0, 0)
    rect(-2, -2, window.width + 2, window.height + 2)

    glColor3f(0, 1, 0)
    rect(1, 1, window.width - 2, window.height - 2)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING
from ctypes import c_void_p, c_uint16, c_uint32, string_at, create_string_buffer, py_object, cast, POINTER
import sys

from pyglet.font.base import GlyphPosition

if TYPE_CHECKING:
    from pyglet.font.freetype import FreeTypeFont
    from pyglet.font.dwrite import Win32DirectWriteFont
    from pyglet.font.quartz import QuartzFont
    from pyglet.font.base import Font, Glyph

from .harfbuzz_lib import hb_lib
_available = False

if hb_lib:
    _available = True

from .harfbuzz_lib import HB_MEMORY_MODE_READONLY, HB_DIRECTION_LTR


def harfbuzz_available() -> bool:
    """Return if the Harfbuzz library is available to use."""
    return _available


_hb_cache: dict[tuple[str, str, bool, str | bool], _HarfbuzzResources] = {}


def _add_utf16_buffer(text: str, buffer: c_void_p) -> None:
    txt_utf16 = text.encode("utf-16-le")  # HarfBuzz expects little-endian UTF-16
    length = len(txt_utf16) // 2
    txt_array = (c_uint16 * length).from_buffer_copy(txt_utf16)
    hb_lib.hb_buffer_add_utf16(buffer, txt_array, length, 0, length)

def _add_utf32_buffer(text: str, buffer: c_void_p) -> None:
    txt_utf32 = text.encode("utf-32-le")
    char_ct = (len(txt_utf32) // 4)
    txt_array = (c_uint32 * char_ct).from_buffer_copy(txt_utf32)
    hb_lib.hb_buffer_add_utf32(buffer, txt_array,char_ct, 0, char_ct)


# Set which unicode python is using for the strings.
if sys.maxunicode == 0x10FFFF:  # UTF-32
    set_buffer_string = _add_utf32_buffer
elif sys.maxunicode == 0xFFFF:  # UTF-16
    set_buffer_string = _add_utf16_buffer

def get_resource_from_ct_font(font: QuartzFont):
    """Get a harfbuzz resource object from a CoreText (Mac) font."""
    key = (font.name, font.weight, font.italic, font.stretch)
    if key in _hb_cache:
        return _hb_cache[key]

    resource = _HarfbuzzResources()
    hb_face = font._get_hb_face()
    resource.load_from_tabled_face(hb_face)
    _hb_cache[key] = resource
    return resource


def get_resource_from_dw_font(font: Win32DirectWriteFont) -> _HarfbuzzResources:
    """Get a harfbuzz resource object from a DirectWrite (Windows) font."""
    key = (font.name, font.weight, font.italic, font.stretch)
    if key in _hb_cache:
        return _hb_cache[key]

    data = font.get_font_data()
    if data:
        resource = _HarfbuzzResources()
        resource.load_from_memory(data)
        _hb_cache[key] = resource
        return resource

    raise Exception("Font data could not be read.")

def get_resource_from_ft_font(font: FreeTypeFont) -> _HarfbuzzResources:
    """Get a harfbuzz resource object from a FreeType (Linux) font."""
    key = (font.name, font.weight, font.italic, font.stretch)
    if key in _hb_cache:
        return _hb_cache[key]

    stream = font.face.ft_face.contents.stream.contents
    data = string_at(stream.base, stream.size)
    resource = _HarfbuzzResources()
    resource.load_from_memory(data)
    _hb_cache[key] = resource
    return resource


def get_resource_from_path(font_path: str):
    with open(font_path, "rb") as f:
        data = f.read()

    resource = _HarfbuzzResources()
    resource.load_from_memory(data)
    return resource

def _get_fallback_glyph_shape(font: Font, text: str) -> tuple[list[Glyph], list[GlyphPosition]]:
    glyphs = []
    offsets = []
    pystr_len = len(text)
    for fallback in font.fallbacks:
        glyph_info = fallback.hb_resource.shape_text(text, fallback.pixel_size)

        indices = [glyph_data["codepoint"] for glyph_data in glyph_info]
        fallback.render_glyph_indices(indices)

        for glyph_data in glyph_info:
            glyph = fallback.glyphs[glyph_data["codepoint"]]
            glyphs.append(glyph)
            offset =  GlyphPosition(glyph_data["x_advance"]-glyph.advance, glyph_data["y_advance"],
                                         glyph_data["x_offset"], glyph_data["y_offset"])
            offsets.append(offset)

    diff = pystr_len - len(glyphs)
    if diff > 0:
        for i in range(diff):
            glyphs.append(font._zero_glyph)
            offsets.append(GlyphPosition(0, 0, 0, 0))

    return glyphs, offsets

def get_harfbuzz_shaped_glyphs(font: Font, text: str) -> tuple[list[Glyph], list[GlyphPosition]]:
    glyph_info = font.hb_resource.shape_text(text, font.pixel_size)
    clusters = [glyph_data["cluster"] for glyph_data in glyph_info]
    cluster_map = defaultdict(list)

    # Find any clusters that have empty indices.
    empty_clusters = {}
    indices = []
    for glyph_data in glyph_info:
        # If any clusters have no codepoint, determine the size to pass to fallback.
        cluster_num = glyph_data["cluster"]
        glyph_index = glyph_data["codepoint"]
        if glyph_index == 0 and cluster_num not in empty_clusters:
            empty_clusters[cluster_num] = clusters.count(cluster_num)

        if cluster_num not in cluster_map:
            cluster_map[cluster_num] = []

        cluster_map[cluster_num].append(glyph_data)
        indices.append(glyph_index)

    font.render_glyph_indices(indices)

    # If a cluster is missing and had 0 indices.
    missing_glyphs = {}
    for cluster_id, cluster_ct in empty_clusters.items():
        missing_text = text[cluster_id:cluster_id + cluster_ct]
        if missing_text not in font.glyphs:
            if missing_text == '\n':
                fb_glyph_info = ([font._zero_glyph], [GlyphPosition(0, 0, 0, 0)])
            else:
                # Get glyphs from fallback, since we need the fallback to actually render it.
                fb_glyph_info = _get_fallback_glyph_shape(font, missing_text)
            font.glyphs[missing_text] = fb_glyph_info  # Cache by string to prevent future fallback paths.
        missing_glyphs[cluster_id] = font.glyphs[missing_text]

    glyphs = []
    offsets = []
    for i in range(len(text)):
        cluster_glyph_info = cluster_map[i]

        # Cluster has data, check if it's part of the missing glyphs.
        if i in missing_glyphs:
            _glyphs, _offsets = missing_glyphs[i]
            glyphs.extend(_glyphs)
            offsets.extend(_offsets)

        # If this cluster is missing, add some zero glyphs until count matches.
        elif len(cluster_glyph_info) == 0:
            while len(glyphs) - 1 < i:
                glyphs.append(font._zero_glyph)
                offsets.append(GlyphPosition(0, 0, 0, 0))
        else:
            for glyph_info in cluster_glyph_info:
                glyph = font.glyphs[glyph_info["codepoint"]]
                glyphs.append(glyph)

                offsets.append(
                    GlyphPosition(glyph_info["x_advance"] - glyph.advance, glyph_info["y_advance"],
                                  glyph_info["x_offset"], glyph_info["y_offset"])
                )

    return glyphs, offsets


class _HarfbuzzResources:
    def __init__(self) -> None:
        self.blob = None
        self.face = None
        self.font = None

    def load_from_tabled_face(self, hb_face: c_void_p) -> None:
        self.blob = None
        self.face = hb_face
        self.font = hb_lib.hb_font_create(self.face)

    def load_from_memory(self, data: bytes) -> None:
        data_len = len(data)
        data_buf = create_string_buffer(data, data_len)
        self.blob = hb_lib.hb_blob_create(data_buf, data_len, HB_MEMORY_MODE_READONLY, None, None)
        self.face = hb_lib.hb_face_create(self.blob, 0)
        self.font = hb_lib.hb_font_create(self.face)

    def shape_text(self, text: str, pixel_size: int, direction: int=HB_DIRECTION_LTR):
        """Shapes the given string using the provided hb_font.
        Returns a list of dictionaries for each glyph containing:
          - codepoint: the glyph index
          - cluster: the index into the original string
          - x_advance, y_advance: the advances (including kerning adjustments)
          - x_offset, y_offset: the offsets (shaped adjustments)
        """
        # Create a new buffer.
        buf = hb_lib.hb_buffer_create()

        # hb_lib.hb_buffer_set_cluster_level(buf, HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS)
        set_buffer_string(text, buf)

        scale = int(pixel_size * 64)
        hb_lib.hb_font_set_scale(self.font, scale, scale)

        hb_lib.hb_buffer_guess_segment_properties(buf)

        # Set text direction (LTR, RTL, etc.).
        hb_lib.hb_buffer_set_direction(buf, direction)

        # Perform shaping.
        hb_lib.hb_shape(self.font, buf, None, 0)

        # Retrieve the number of glyphs.
        length = hb_lib.hb_buffer_get_length(buf)

        # Get pointers to the glyph info and position arrays.
        infos = hb_lib.hb_buffer_get_glyph_infos(buf, None)
        positions = hb_lib.hb_buffer_get_glyph_positions(buf, None)

        # Collect glyph metrics.
        glyphs = []
        for i in range(length):
            info = infos[i]
            pos = positions[i]
            glyphs.append({
                "codepoint": info.codepoint,
                "cluster": info.cluster,
                "x_advance": pos.x_advance / 64.0,
                "y_advance": pos.y_advance / 64.0,
                "x_offset": pos.x_offset / 64.0,
                "y_offset": pos.y_offset / 64.0,
            })

        # Clean up the buffer.
        hb_lib.hb_buffer_destroy(buf)

        return glyphs

    def __del__(self):
        self.destroy()

    def destroy(self):
        if self.font:
            hb_lib.hb_font_destroy(self.font)
            self.font = None

        if self.face:
            hb_lib.hb_face_destroy(self.face)
            self.face = None

        if self.blob:
            hb_lib.hb_blob_destroy(self.blob)
            self.blob = None



#!/usr/bin/env python
"""Print OpenAL driver information.

Options:
  -d <device>   Optionally specify device to query.
"""
import ctypes
import optparse
import sys

from pyglet.util import asbytes
from pyglet.media.drivers import openal
from pyglet.media.drivers.openal.interface import alc


def split_nul_strings(s):
    # NUL-separated list of strings, double-NUL-terminated.
    nul = False
    i = 0
    while True:
        if s[i] == b'\x00':
            if nul:
                break
            else:
                nul = True
        else:
            nul = False
        i += 1
    s = s[:i - 1]
    return s.split(b'\x00')


if __name__ == '__main__':
    op = optparse.OptionParser()
    op.add_option('-d', '--device', dest='device',
                  help='use device DEVICE', metavar='DEVICE')
    (options, args) = op.parse_args(sys.argv[1:])

    default_device = ctypes.cast(
        alc.alcGetString(None, alc.ALC_DEFAULT_DEVICE_SPECIFIER), ctypes.c_char_p).value
    capture_default_device = ctypes.cast(
        alc.alcGetString(None, alc.ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER), ctypes.c_char_p).value

    print('Default device:         %s' % default_device)
    print('Default capture device: %s' % capture_default_device)

    if alc.alcIsExtensionPresent(None, ctypes.create_string_buffer(b'ALC_ENUMERATION_EXT')):
        devices = split_nul_strings(alc.alcGetString(None, alc.ALC_DEVICE_SPECIFIER))
        capture_devices = split_nul_strings(alc.alcGetString(None, alc.ALC_CAPTURE_DEVICE_SPECIFIER))

        print('Devices:                %s' % b', '.join(devices))
        print('Capture devices:        %s\n' % b', '.join(capture_devices))

    if options.device:
        print('Using device "%s"...' % options.device)
        driver = openal.create_audio_driver(asbytes(options.device))
    else:
        print('Using default device...')
        driver = openal.create_audio_driver()

    print('OpenAL version %d.%d' % driver.get_version())
    print('Extensions:              %s' % ', '.join(driver.get_extensions()))



# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
# $Id$

import os

from pyglet import media


class ReaderException(Exception):
    pass


class PlayerReader:
    def __init__(self, player):
        self.player = player

    def line(self, line, lineno):
        parts = line.split()
        if parts[0] == 'position':
            if len(parts) < 4:
                raise ReaderException('Invalid position line %d' % lineno)
            self.player.position = tuple([float(x) for x in parts[1:]])
        if parts[0] == 'cone_orientation':
            if len(parts) < 4:
                raise ReaderException('Invalid orientation line %d' % lineno)
            self.player.cone_orientation = tuple([float(x) for x in parts[1:]])
        elif parts[0] == 'outer_cone_angle':
            if len(parts) < 2:
                raise ReaderException('Invalid angle line %d' % lineno)
            self.player.cone_outer_angle = float(parts[1])
        elif parts[0] == 'inner_cone_angle':
            if len(parts) < 2:
                raise ReaderException('Invalid angle line %d' % lineno)
            self.player.cone_inner_angle = float(parts[1])
        elif parts[0] == 'label':
            if len(parts) < 2:
                raise ReaderException('Invalid label line %d' % lineno)
            self.player.label = parts[1]


class SpaceReader:
    def __init__(self, space):
        self.basedir = ''
        self.space = space

    def read(self, file):
        if not hasattr(file, 'read'):
            self.basedir = os.path.dirname(file)
            with open(file) as f:
                file = f.readlines()
        elif hasattr(file, 'name'):
            self.basedir = os.path.dirname(file.name)
        reader = None
        player = None
        lineno = 0
        for line in file:
            lineno += 1

            if not line.strip() or line.startswith('#'):
                continue

            if line.startswith(' '):
                if not reader:
                    raise ReaderException('Unexpected indented block line %d' % lineno)
                reader.line(line, lineno)
            else:
                reader = None
                if player is not None:
                    self.space.add_player(player)
                    player = None

                parts = line.split()
                if parts[0] == 'loop':
                    if len(parts) < 2:
                        raise ReaderException('No loop filename line %d' % lineno)
                    player = media.Player()
                    player.loop = True
                    player.queue(self.source(parts[1], streaming=False))
                    reader = PlayerReader(player)

        if player is not None:
            self.space.add_player(player)

    def source(self, filename, **kwargs):
        filename = os.path.join(self.basedir, filename)
        return media.load(filename, **kwargs)



from __future__ import annotations

import os
import time
import ctypes
import warnings
import threading

from ctypes import c_uint16 as _u16
from ctypes import c_int16 as _s16
from ctypes import c_uint32 as _u32
from ctypes import c_int32 as _s32
from ctypes import c_int64 as _s64

import pyglet

from .evdev_constants import *
from pyglet.app.xlib import XlibSelectDevice
from pyglet.libs.ioctl import _IOR, _IOR_str, _IOR_len, _IOW
from pyglet.input.base import Device, RelativeAxis, AbsoluteAxis, Button, Joystick, Controller
from pyglet.input.base import DeviceOpenException, ControllerManager
from pyglet.input.controller import get_mapping, Relation, create_guid

try:
    from os import readv as _readv
except ImportError:
    # Workaround for missing os.readv in PyPy
    c = pyglet.lib.load_library('c')

    def _readv(fd, buffers):
        return c.read(fd, buffers, 3072)


# Structures from /linux/blob/master/include/uapi/linux/input.h

class Timeval(ctypes.Structure):
    _fields_ = (
        ('tv_sec', _s64),
        ('tv_usec', _s64)
    )


class InputEvent(ctypes.Structure):
    _fields_ = (
        ('time', Timeval),
        ('type', _u16),
        ('code', _u16),
        ('value', _s32)
    )


class InputID(ctypes.Structure):
    _fields_ = (
        ('bustype', _u16),
        ('vendor', _u16),
        ('product', _u16),
        ('version', _u16),
    )


class InputABSInfo(ctypes.Structure):
    _fields_ = (
        ('value', _s32),
        ('minimum', _s32),
        ('maximum', _s32),
        ('fuzz', _s32),
        ('flat', _s32),
    )


class FFReplay(ctypes.Structure):
    _fields_ = (
        ('length', _u16),
        ('delay', _u16)
    )


class FFTrigger(ctypes.Structure):
    _fields_ = (
        ('button', _u16),
        ('interval', _u16)
    )


class FFEnvelope(ctypes.Structure):
    _fields_ = [
        ('attack_length', _u16),
        ('attack_level', _u16),
        ('fade_length', _u16),
        ('fade_level', _u16),
    ]


class FFConstantEffect(ctypes.Structure):
    _fields_ = [
        ('level', _s16),
        ('ff_envelope', FFEnvelope),
    ]


class FFRampEffect(ctypes.Structure):
    _fields_ = [
        ('start_level', _s16),
        ('end_level', _s16),
        ('ff_envelope', FFEnvelope),
    ]


class FFConditionEffect(ctypes.Structure):
    _fields_ = [
        ('right_saturation', _u16),
        ('left_saturation', _u16),
        ('right_coeff', _s16),
        ('left_coeff', _s16),
        ('deadband', _u16),
        ('center', _s16),
    ]


class FFPeriodicEffect(ctypes.Structure):
    _fields_ = [
        ('waveform', _u16),
        ('period', _u16),
        ('magnitude', _s16),
        ('offset', _s16),
        ('phase', _u16),
        ('envelope', FFEnvelope),
        ('custom_len', _u32),
        ('custom_data', ctypes.POINTER(_s16)),
    ]


class FFRumbleEffect(ctypes.Structure):
    _fields_ = (
        ('strong_magnitude', _u16),
        ('weak_magnitude', _u16)
    )


class FFEffectType(ctypes.Union):
    _fields_ = (
        ('ff_constant_effect', FFConstantEffect),
        ('ff_ramp_effect', FFRampEffect),
        ('ff_periodic_effect', FFPeriodicEffect),
        ('ff_condition_effect', FFConditionEffect * 2),
        ('ff_rumble_effect', FFRumbleEffect),
    )


class FFEvent(ctypes.Structure):
    _fields_ = (
        ('type', _u16),
        ('id', _s16),
        ('direction', _u16),
        ('ff_trigger', FFTrigger),
        ('ff_replay', FFReplay),
        ('u', FFEffectType)
    )


EVIOCGVERSION = _IOR('E', 0x01, ctypes.c_int)
EVIOCGID = _IOR('E', 0x02, InputID)
EVIOCGNAME = _IOR_str('E', 0x06)
EVIOCGPHYS = _IOR_str('E', 0x07)
EVIOCGUNIQ = _IOR_str('E', 0x08)
EVIOCSFF = _IOW('E', 0x80, FFEvent)


def EVIOCGBIT(fileno, ev, buffer):
    return _IOR_len('E', 0x20 + ev)(fileno, buffer)


def EVIOCGABS(fileno, abs):
    buffer = InputABSInfo()
    return _IOR_len('E', 0x40 + abs)(fileno, buffer)


def get_set_bits(bytestring):
    bits = set()
    j = 0
    for byte in bytestring:
        for i in range(8):
            if byte & 1:
                bits.add(j + i)
            byte >>= 1
        j += 8
    return bits


_abs_names = {
    ABS_X: AbsoluteAxis.X,
    ABS_Y: AbsoluteAxis.Y,
    ABS_Z: AbsoluteAxis.Z,
    ABS_RX: AbsoluteAxis.RX,
    ABS_RY: AbsoluteAxis.RY,
    ABS_RZ: AbsoluteAxis.RZ,
    ABS_HAT0X: AbsoluteAxis.HAT_X,
    ABS_HAT0Y: AbsoluteAxis.HAT_Y,
}

_rel_names = {
    REL_X: RelativeAxis.X,
    REL_Y: RelativeAxis.Y,
    REL_Z: RelativeAxis.Z,
    REL_RX: RelativeAxis.RX,
    REL_RY: RelativeAxis.RY,
    REL_RZ: RelativeAxis.RZ,
    REL_WHEEL: RelativeAxis.WHEEL,
}


def _create_control(fileno, event_type, event_code):
    if event_type == EV_ABS:
        raw_name = abs_raw_names.get(event_code, f'EV_ABS({event_code:x})')
        name = _abs_names.get(event_code)
        absinfo = EVIOCGABS(fileno, event_code)
        value = absinfo.value
        minimum = absinfo.minimum
        maximum = absinfo.maximum
        control = AbsoluteAxis(name, minimum, maximum, raw_name)
        control.value = value
        if name == 'hat_y':
            control.inverted = True
    elif event_type == EV_REL:
        raw_name = rel_raw_names.get(event_code, f'EV_REL({event_code:x})')
        name = _rel_names.get(event_code)
        control = RelativeAxis(name, raw_name)
    elif event_type == EV_KEY:
        raw_name = key_raw_names.get(event_code, f'EV_KEY({event_code:x})')
        name = None
        control = Button(name, raw_name)
    else:
        return None
    control.event_type = event_type
    control.event_code = event_code
    return control


event_types = {
    EV_KEY: KEY_MAX,
    EV_REL: REL_MAX,
    EV_ABS: ABS_MAX,
    EV_MSC: MSC_MAX,
    EV_LED: LED_MAX,
    EV_SND: SND_MAX,
    EV_FF: FF_MAX,
}


class EvdevDevice(XlibSelectDevice, Device):
    _fileno = None

    def __init__(self, display, filename):
        self._filename = filename

        fileno = os.open(filename, os.O_RDONLY)

        self._id = EVIOCGID(fileno)
        self.id_bustype = self._id.bustype
        self.id_vendor = hex(self._id.vendor)
        self.id_product = hex(self._id.product)
        self.id_version = self._id.version

        name = EVIOCGNAME(fileno)
        try:
            name = name.decode('utf-8')
        except UnicodeDecodeError:
            try:
                name = name.decode('latin-1')
            except UnicodeDecodeError:
                pass

        try:
            self.phys = EVIOCGPHYS(fileno)
        except OSError:
            self.phys = ''
        try:
            self.uniq = EVIOCGUNIQ(fileno)
        except OSError:
            self.uniq = ''

        self.controls = []
        self.control_map = {}
        self.ff_types = []

        event_types_bits = (ctypes.c_byte * 4)()
        EVIOCGBIT(fileno, 0, event_types_bits)
        for event_type in get_set_bits(event_types_bits):
            if event_type not in event_types:
                continue
            max_code = event_types[event_type]
            nbytes = max_code // 8 + 1
            event_codes_bits = (ctypes.c_byte * nbytes)()
            EVIOCGBIT(fileno, event_type, event_codes_bits)
            if event_type == EV_FF:
                self.ff_types.extend(get_set_bits(event_codes_bits))
            else:
                for event_code in get_set_bits(event_codes_bits):
                    control = _create_control(fileno, event_type, event_code)
                    if control:
                        self.control_map[(event_type, event_code)] = control
                        self.controls.append(control)

        self.controls.sort(key=lambda c: c.event_code)
        os.close(fileno)

        self._event_size = ctypes.sizeof(InputEvent)
        self._event_buffer = (InputEvent * 64)()

        super().__init__(display, name)

    @property
    def connected(self):
        return os.path.exists(self._filename)

    def get_guid(self):
        """Get the device's SDL2 style GUID string"""
        _id = self._id
        return create_guid(_id.bustype, _id.vendor, _id.product, _id.version, self.name, 0, 0)

    def open(self, window=None, exclusive=False):
        try:
            self._fileno = os.open(self._filename, os.O_RDWR | os.O_NONBLOCK)
        except OSError as e:
            raise DeviceOpenException(e)

        pyglet.app.platform_event_loop.select_devices.add(self)
        super().open(window, exclusive)

    def close(self):
        super().close()

        if not self._fileno:
            return

        pyglet.app.platform_event_loop.select_devices.remove(self)
        os.close(self._fileno)
        self._fileno = None

    def get_controls(self):
        return self.controls

    # Force Feedback methods

    def ff_upload_effect(self, structure):
        os.write(self._fileno, structure)

    # XlibSelectDevice interface

    def fileno(self):
        return self._fileno

    def poll(self):
        return False

    def select(self):
        if not self._fileno:
            return

        try:
            bytes_read = _readv(self._fileno, self._event_buffer)
        except OSError:
            self.close()
            return

        n_events = bytes_read // self._event_size

        for event in self._event_buffer[:n_events]:
            try:
                self.control_map[(event.type, event.code)].value = event.value
            except KeyError:
                pass


class FFController(Controller):
    """Controller that supports force-feedback"""
    _fileno = None
    _weak_effect = None
    _play_weak_event = None
    _stop_weak_event = None
    _strong_effect = None
    _play_strong_event = None
    _stop_strong_event = None

    device: EvdevDevice

    def open(self, window=None, exclusive=False):
        super().open(window, exclusive)
        self._fileno = self.device.fileno()
        # Create Force Feedback effects & events when opened:
        # https://www.kernel.org/doc/html/latest/input/ff.html
        self._weak_effect = FFEvent(FF_RUMBLE, -1)
        EVIOCSFF(self._fileno, self._weak_effect)
        self._play_weak_event = InputEvent(Timeval(), EV_FF, self._weak_effect.id, 1)
        self._stop_weak_event = InputEvent(Timeval(), EV_FF, self._weak_effect.id, 0)

        self._strong_effect = FFEvent(FF_RUMBLE, -1)
        EVIOCSFF(self._fileno, self._strong_effect)
        self._play_strong_event = InputEvent(Timeval(), EV_FF, self._strong_effect.id, 1)
        self._stop_strong_event = InputEvent(Timeval(), EV_FF, self._strong_effect.id, 0)

    def rumble_play_weak(self, strength=1.0, duration=0.5):
        assert self._fileno, "Controller must be opened first."
        effect = self._weak_effect
        effect.u.ff_rumble_effect.weak_magnitude = int(max(min(1.0, strength), 0) * 0xFFFF)
        effect.ff_replay.length = int(duration * 1000)
        EVIOCSFF(self._fileno, effect)
        self.device.ff_upload_effect(self._play_weak_event)

    def rumble_play_strong(self, strength=1.0, duration=0.5):
        assert self._fileno, "Controller must be opened first."
        effect = self._strong_effect
        effect.u.ff_rumble_effect.strong_magnitude = int(max(min(1.0, strength), 0) * 0xFFFF)
        effect.ff_replay.length = int(duration * 1000)
        EVIOCSFF(self._fileno, effect)
        self.device.ff_upload_effect(self._play_strong_event)

    def rumble_stop_weak(self):
        assert self._fileno, "Controller must be opened first."
        self.device.ff_upload_effect(self._stop_weak_event)

    def rumble_stop_strong(self):
        assert self._fileno, "Controller must be opened first."
        self.device.ff_upload_effect(self._stop_strong_event)


class EvdevControllerManager(ControllerManager, XlibSelectDevice):

    def __init__(self, display=None):
        super().__init__()
        self._display = display
        self._devices_file = open('/proc/bus/input/devices')
        self._device_names = self._get_device_names()
        self._controllers = {}

        for name in self._device_names:
            try:
                path = os.path.join('/dev/input', name)
                device = EvdevDevice(self._display, path)
            except OSError:
                continue
            if controller := _create_controller(device):
                self._controllers[name] = controller

        pyglet.app.platform_event_loop.select_devices.add(self)

    def __del__(self):
        self._devices_file.close()

    def fileno(self):
        """Allow this class to be Selectable"""
        return self._devices_file.fileno()

    @staticmethod
    def _get_device_names():
        return {name for name in os.listdir('/dev/input') if name.startswith('event')}

    def _make_controller_thread(self, name: str, retries: int) -> None:
        # Try to create a Device:
        for _ in range(retries):
            try:
                path = os.path.join('/dev/input', name)
                device = EvdevDevice(self._display, path)
                break
            except OSError:
                time.sleep(0.1)
        else:
            return  # No device could be created

        # Reuse existing controller instance if it exists, or create a new one:
        if controller := self._controllers.get(name, _create_controller(device)):
            self._controllers[name] = controller
            # Dispatch event in main thread:
            self.post_event('on_connect', controller)

    def select(self):
        """Triggered whenever the devices_file changes."""
        new_device_files = self._get_device_names()
        appeared = new_device_files - self._device_names
        disappeared = self._device_names - new_device_files
        self._device_names = new_device_files

        for name in appeared:
            t = threading.Thread(target=self._make_controller_thread, args=(name, 10), daemon=True)
            t.start()

        for name in disappeared:
            if controller := self._controllers.get(name):
                self.dispatch_event('on_disconnect', controller)

    def get_controllers(self) -> list[Controller]:
        return [_c for _c in self._controllers.values() if _c.device.connected is True]


def get_devices(display=None):
    _devices = {}
    base = '/dev/input'
    for filename in os.listdir(base):
        if filename.startswith('event'):
            path = os.path.join(base, filename)
            if path in _devices:
                continue

            try:
                _devices[path] = EvdevDevice(display, path)
            except OSError:
                pass

    return list(_devices.values())


def _create_joystick(device):
    # Look for something with an ABS X and ABS Y axis, and a joystick 0 button
    have_x = False
    have_y = False
    have_button = False
    for control in device.controls:
        if control.event_type == EV_ABS and control.event_code == ABS_X:
            have_x = True
        elif control.event_type == EV_ABS and control.event_code == ABS_Y:
            have_y = True
        elif control.event_type == EV_KEY and control.event_code in (BTN_JOYSTICK, BTN_GAMEPAD):
            have_button = True
    if not (have_x and have_y and have_button):
        return

    return Joystick(device)


def get_joysticks(display=None):
    return [joystick for joystick in [_create_joystick(device) for device in get_devices(display)] if joystick]


def _detect_controller_mapping(device):
    # If no explicit mapping is available, we can
    # detect it from the Linux gamepad specification:
    # https://www.kernel.org/doc/html/latest/input/gamepad.html
    # Note: legacy device drivers don't always adhere to this.
    mapping = dict(guid=device.get_guid(), name=device.name)

    _aliases = {BTN_MODE: 'guide', BTN_SELECT: 'back', BTN_START: 'start',
                BTN_SOUTH: 'a', BTN_EAST: 'b', BTN_WEST: 'x', BTN_NORTH: 'y',
                BTN_TL: 'leftshoulder', BTN_TR: 'rightshoulder',
                BTN_TL2: 'lefttrigger', BTN_TR2: 'righttrigger',
                BTN_THUMBL: 'leftstick', BTN_THUMBR: 'rightstick',
                BTN_DPAD_UP: 'dpup', BTN_DPAD_DOWN: 'dpdown',
                BTN_DPAD_LEFT: 'dpleft', BTN_DPAD_RIGHT: 'dpright',

                ABS_HAT0X: 'dpleft',  # and 'dpright',
                ABS_HAT0Y: 'dpup',    # and 'dpdown',
                ABS_Z: 'lefttrigger', ABS_RZ: 'righttrigger',
                ABS_X: 'leftx', ABS_Y: 'lefty', ABS_RX: 'rightx', ABS_RY: 'righty'}

    button_controls = [control for control in device.controls if isinstance(control, Button)]
    axis_controls = [control for control in device.controls if isinstance(control, AbsoluteAxis)]
    hat_controls = [control for control in device.controls if control.name in ('hat_x', 'hat_y')]

    for i, control in enumerate(button_controls):
        if name := _aliases.get(control.event_code):
            mapping[name] = Relation('button', i)

    for i, control in enumerate(axis_controls):
        if name := _aliases.get(control.event_code):
            mapping[name] = Relation('axis', i)

    for i, control in enumerate(hat_controls):
        if name := _aliases.get(control.event_code):
            index = 1 + i << 1
            mapping[name] = Relation('hat0', index)

    return mapping


def _create_controller(device) -> Controller | None:
    for control in device.controls:
        if control.event_type == EV_KEY and control.event_code == BTN_GAMEPAD:
            break
    else:
        return None     # Game Controllers must have a BTN_GAMEPAD

    mapping = get_mapping(device.get_guid())
    if not mapping:
        warnings.warn(f"Warning: {device} (GUID: {device.get_guid()}) "
                      f"has no controller mappings. Update the mappings in the Controller DB.\n"
                      f"Auto-detecting as defined by the 'Linux gamepad specification'")
        mapping = _detect_controller_mapping(device)

    if FF_RUMBLE in device.ff_types:
        return FFController(device, mapping)
    else:
        return Controller(device, mapping)


def get_controllers(display=None):
    return [controller for controller in [_create_controller(device) for device in get_devices(display)] if controller]



import pyglet

"""Camera class for easy scrolling and zooming.

A simple example of a Camera class that can be used to easily scroll and
zoom when rendering. For example, you might have a playfield that needs  
to scroll and/or zoom, and a GUI layer that will remain static. For that
scenario, you can create two Camera instances. You can optionally set
the minimum allowed zoom, maximum allowed zoom, and scrolling speed::

    world_camera = Camera(scroll_speed=5, min_zoom=1, max_zoom=4)
    gui_camera = Camera()

After creating Camera instances, the zoom can be easily updated. It will
clamp to the `max_zoom` parameter (default of 4)::

    world_camera.zoom += 1

The scrolling can be set in two different ways. Directly with the
`Camera.position attribute, which can be set with a tuple of absolute
x, y values::

    world_camera.position = 50, 0

Or, it can be updated incrementally with the `Camera.move(x, y)` method.
This will update the camera position by multiplying the passed vector by
the `Camera.scroll_speed` parameter, which can be set on instantiation. 

    world_camera.move(1, 0)
    # If the world_camera.scroll_speed is "5", this will move the camera
    # by 5 pixels right on the x axis. 


During your `Window.on_draw` event, you can set the Camera, and draw the
appropriate objects. For convenience, the Camera class can act as a context
manager, allowing easy use of "with"::

    @window.event
    def on_draw():
        window.clear()
    
        # Draw your world scene using the world camera
        with world_camera:
            batch.draw()
    
        # Can also be written as:
        # camera.begin()
        # batch.draw()
        # camera.end()
    
        # Draw your GUI elements with the GUI camera.
        with gui_camera:
            label.draw()

"""


class Camera:
    """ A simple 2D camera that contains the speed and offset."""

    def __init__(self, window: pyglet.window.Window, scroll_speed=1, min_zoom=1, max_zoom=4):
        assert min_zoom <= max_zoom, "Minimum zoom must not be greater than maximum zoom"
        self._window = window
        self.scroll_speed = scroll_speed
        self.max_zoom = max_zoom
        self.min_zoom = min_zoom
        self.offset_x = 0
        self.offset_y = 0
        self._zoom = max(min(1, self.max_zoom), self.min_zoom)

    @property
    def zoom(self):
        return self._zoom

    @zoom.setter
    def zoom(self, value):
        """ Here we set zoom, clamp value to minimum of min_zoom and max of max_zoom."""
        self._zoom = max(min(value, self.max_zoom), self.min_zoom)

    @property
    def position(self):
        """Query the current offset."""
        return self.offset_x, self.offset_y

    @position.setter
    def position(self, value):
        """Set the scroll offset directly."""
        self.offset_x, self.offset_y = value

    def move(self, axis_x, axis_y):
        """ Move axis direction with scroll_speed.
            Example: Move left -> move(-1, 0)
         """
        self.offset_x += self.scroll_speed * axis_x
        self.offset_y += self.scroll_speed * axis_y

    def begin(self):
        # Set the current camera offset so you can draw your scene.

        # Translate using the offset.
        view_matrix = self._window.view.translate((-self.offset_x * self._zoom, -self.offset_y * self._zoom, 0))
        # Scale by zoom level.
        view_matrix = view_matrix.scale((self._zoom, self._zoom, 1))

        self._window.view = view_matrix

    def end(self):
        # Since this is a matrix transformation, you need to reverse the scaling and translation
        # after rendering. Otherwise, the transformations will accumulate with each draw call,
        # pushing the view further and further away.
        
        # Reverse the scaling since it was the last transformation.
        view_matrix = self._window.view.scale((1 / self._zoom, 1 / self._zoom, 1))
        # Reverse the translation.
        view_matrix = view_matrix.translate((self.offset_x * self._zoom, self.offset_y * self._zoom, 0))

        self._window.view = view_matrix

    def __enter__(self):
        self.begin()

    def __exit__(self, exception_type, exception_value, traceback):
        self.end()


class CenteredCamera(Camera):
    """A simple 2D camera class. 0, 0 will be the center of the screen, as opposed to the bottom left."""

    def begin(self):
        x = -self._window.width // 2 / self._zoom + self.offset_x
        y = -self._window.height // 2 / self._zoom + self.offset_y

        view_matrix = self._window.view.translate((-x * self._zoom, -y * self._zoom, 0))
        view_matrix = view_matrix.scale((self._zoom, self._zoom, 1))
        self._window.view = view_matrix

    def end(self):
        x = -self._window.width // 2 / self._zoom + self.offset_x
        y = -self._window.height // 2 / self._zoom + self.offset_y

        view_matrix = self._window.view.scale((1 / self._zoom, 1 / self._zoom, 1))
        view_matrix = view_matrix.translate((x * self._zoom, y * self._zoom, 0))
        self._window.view = view_matrix




import pyglet

"""Simple example demonstrating how to get the text in the clipboard as well as putting text in the clipboard."""

window = pyglet.window.Window()
print("Existing Clipboard Text:", window.get_clipboard_text())


@window.event
def on_key_press(symbol, modifiers):
    print("Clipboard changed to: Hello World!")

    window.set_clipboard_text("Hello World!")

    print("Clipboard Text:", window.get_clipboard_text())


@window.event
def on_draw():
    window.clear()


pyglet.app.run()



import pyglet

window = pyglet.window.Window(resizable=True)

# Set example resource path.
pyglet.resource.path = ['../resources']
pyglet.resource.reindex()

# Load example image from resource path.
image = pyglet.resource.image("pyglet.png")

# Anchor point on an image is bottom left corner by default.
# Set to center point with anchor properties.
image.anchor_x = image.width // 2
image.anchor_y = image.height // 2

# Create basic sprite in the center of the window.
sprite = pyglet.sprite.Sprite(image, x=window.width // 2, y=window.height // 2)

@window.event
def on_resize(width, height):
    # This will keep sprite at the center of the window as it changes size.
    sprite.update(x=width // 2, y=height // 2)

@window.event
def on_draw():
    window.clear()
    sprite.draw()

pyglet.app.run()



from __future__ import annotations

import ctypes
from ctypes import CFUNCTYPE, POINTER, c_char_p, c_int, cast
from typing import Any, Callable, Sequence

import pyglet
from pyglet.gl.lib import decorate_function, missing_function
from pyglet.util import asbytes

__all__ = ['link_GL', 'link_WGL']

_debug_trace = pyglet.options['debug_trace']

gl_lib = ctypes.windll.opengl32
wgl_lib = gl_lib

if _debug_trace:
    from pyglet.lib import _TraceLibrary

    gl_lib = _TraceLibrary(gl_lib)
    wgl_lib = _TraceLibrary(wgl_lib)

try:
    wglGetProcAddress = wgl_lib.wglGetProcAddress  # noqa: N816
    wglGetProcAddress.restype = CFUNCTYPE(POINTER(c_int))
    wglGetProcAddress.argtypes = [c_char_p]
    _have_get_proc_address = True
except AttributeError:
    _have_get_proc_address = False

class_slots = ['name', 'requires', 'suggestions', 'ftype', 'func']


def makeWGLFunction(func: Callable) -> Callable:  # noqa: N802
    class WGLFunction:
        __slots__ = class_slots
        __call__ = func

    return WGLFunction


class WGLFunctionProxy:
    __slots__ = class_slots

    def __init__(self, name: str, ftype: ctypes.WINFUNCTYPE, requires: str | None,
                 suggestions: list[str] | None) -> None:
        assert _have_get_proc_address
        self.name = name
        self.ftype = ftype
        self.requires = requires
        self.suggestions = suggestions
        self.func = None

    def __call__(self, *args: Any, **kwargs: Any) -> Callable:
        from pyglet.gl import current_context
        if not current_context:
            msg = f'Call to function "{self.name}" before GL context created'
            raise Exception(msg)
        address = wglGetProcAddress(asbytes(self.name))
        if cast(address, POINTER(c_int)):  # check cast because address is func
            self.func = cast(address, self.ftype)
            decorate_function(self.func, self.name)
        else:
            self.func = missing_function(
                self.name, self.requires, self.suggestions)

        self.__class__ = makeWGLFunction(self.func)

        return self.func(*args, **kwargs)


def link_GL(name: str, restype: Any, argtypes: Any, requires: str | None = None,  # noqa: N802, D103
            suggestions: Sequence[str] | None = None) -> Callable[..., Any]:
    try:
        func = getattr(gl_lib, name)
        func.restype = restype
        func.argtypes = argtypes
        decorate_function(func, name)
        return func
    except AttributeError:
        # Not in opengl32.dll. Try and get a pointer from WGL.
        try:
            fargs = (restype,) + tuple(argtypes)
            ftype = ctypes.WINFUNCTYPE(*fargs)
            if _have_get_proc_address:
                from pyglet.gl import gl_info
                if gl_info.have_context():
                    address = wglGetProcAddress(name)
                    if address:
                        func = cast(address, ftype)
                        decorate_function(func, name)
                        return func
                else:
                    # Insert proxy until we have a context
                    return WGLFunctionProxy(name, ftype, requires, suggestions)
        except:  # noqa: E722, S110
            # TODO: Figure out what exception this can cause instead of catching all.
            pass

        return missing_function(name, requires, suggestions)


link_WGL = link_GL  # noqa: N816



"""This module defines the usage and creation of user defined fonts in Pyglet.

Previously, pyglet only supported font renderers that are built into the operating system, such as
``FreeType``, ``DirectWrite``, or ``Quartz``. However, there are situations in which a user may not want or need all the
features a font can provide. They just need to put characters in a particular order without the hassle of exporting
into a separate file.

The :py:class:`~pyglet.font.user.UserDefinedMappingFont` is provided for most use cases, which will allow you to
make an internal font that can be used where a ``font_name`` is required to identify a font.

A user defined font is also identified by its name. The name you choose should be unique to ensure it will not conflict
with a system font. For example, do not use `Arial`, as that will collide with Windows systems.

With :py:class:`~pyglet.font.user.UserDefinedMappingFont` you can pass a mapping of characters that point to your
:py:class:`~pyglet.image.ImageData`.

.. code-block:: python

    mappings={'c': my_image_data, 'b': my_image_data, 'a': my_image_data}

For more custom behavior, a dict-like object can be used, such as a class.

.. code-block:: python

    class MyCustomMapping:
        def get(self, char: str) -> ImageData | None:
            # return ImageData if a character is found.
            # return None if no character is found

    mappings = MyCustomMapping()

Once your font is created, you also must register it within pyglet to use it. This can be done through the
 :py:func:`~pyglet.font.add_user_font` function.

When you register a user defined font, only those parameters will used to identify the font. If you have a font, but
want to have a ``italic`` enabled version. You must make a new instance of your font, but with the ``italic``
parameter set as ``True``. Same applies to the ``size`` parameter. The ``weight`` parameter can also be provided as
a string.

Scaling
=======
By default, user font's will not be scaled. In most use cases, you have a single font at a specific size that you
want to use.

There are cases where a user may want to scale their font to be used at any size. We provide the following function:
:py:func:`~pyglet.font.user.get_scaled_user_font`. By providing the user defined font instance, and a new size, you will
get back a new font instance that is scaled to the new size. This new instance must also be registered the same way as
the base font.

When specifying the ``size`` parameter, that value is used to determine the ratio of scaling between the new size. So
if your base font is a size of 12, creating a scaled version at size 24 will be double the size of the base.

.. warning::

    The ``PIL`` library is a required dependency to use the scaling functionality.

.. versionadded:: 2.0.15
"""
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Protocol

import pyglet
from pyglet.font import base

SCALING_ENABLED = False
try:
    from PIL import Image
    from PIL.Image import Resampling

    SCALING_ENABLED = True
except ImportError:
    pass

if TYPE_CHECKING:
    from pyglet.font.base import Glyph, GlyphPosition
    from pyglet.image import ImageData


class UserDefinedGlyphRenderer(base.GlyphRenderer):
    def __init__(self, font: UserDefinedFontBase) -> None:
        super().__init__(font)
        self._font = font

    def render(self, image_data: ImageData) -> Glyph:
        if self._font._scaling:  # noqa: SLF001
            image_original = Image.frombytes("RGBA", (image_data.width, image_data.height),
                                             image_data.get_image_data().get_data("RGBA"))
            scale_ratio = self._font.size / self._font._base_size
            image_resized = image_original.resize((int(image_data.width * scale_ratio),
                                                   int(image_data.height * scale_ratio)), Resampling.NEAREST)
            new_image = pyglet.image.ImageData(image_resized.width, image_resized.height,
                                               "RGBA", image_resized.tobytes())
            glyph = self._font.create_glyph(new_image)
            glyph.set_bearings(-self._font.descent, 0, image_resized.width)
        else:
            glyph = self._font.create_glyph(image_data)
            glyph.set_bearings(-self._font.descent, 0, image_data.width)
        return glyph


class UserDefinedFontBase(base.Font):
    """Used as a base for all user defined fonts.

    .. versionadded:: 2.0.15
    """
    glyph_renderer_class: ClassVar[type[base.GlyphRenderer]] = UserDefinedGlyphRenderer

    def __init__(
            self, name: str, default_char: str, size: int, ascent: int | None = None, descent: int | None = None,
            weight: str = "normal", italic: bool = False, stretch: bool = False, dpi: int = 96, locale: str | None = None,
    ) -> None:
        """Initialize a user defined font.

        Args:
            name:
                Name of the font. Used to identify the font. Must be unique to ensure it does not
                collide with any system fonts.
            default_char:
                If a character in a string is not found in the font, it will use this as fallback.
            size:
                Font size, usually in pixels.
            ascent:
                Maximum ascent above the baseline, in pixels. If None, the image height is used.
            descent:
                Maximum descent below the baseline, in pixels. Usually negative.
            weight:
                The font weight, as a string. Defaults to "normal".
            italic:
                If True, this font will be used when ``italic`` is enabled for the font name.
            stretch:
                If True, this font will be used when ``stretch`` is enabled for the font name.
            dpi:
                The assumed resolution of the display device, for the purposes of determining the pixel size of the
                font. Use a default of 96 for standard sizing.
            locale:
                Used to specify the locale of this font.
        """
        super().__init__()
        self._name = name
        self.default_char = default_char
        self.ascent = ascent
        self.descent = descent
        self.size = size
        self.weight = weight
        self.italic = italic
        self.stretch = stretch
        self.dpi = dpi
        self.locale = locale

        self._base_size = 0
        self._scaling = False

    @property
    def name(self) -> str:
        return self._name

    def enable_scaling(self, base_size: int) -> None:
        if not SCALING_ENABLED:
            msg = "PIL is not installed. User Font Scaling requires PIL."
            raise ImportError(msg)

        self._base_size = base_size
        self._scaling = True

    def _initialize_renderer(self) -> None:
        """Initialize the glyph renderer and cache it on the Font.

        This way renderers for fonts that have been loaded but not used will not have unnecessary loaders.
        """
        if not self._glyph_renderer:
            self._glyph_renderer = self.glyph_renderer_class(self)

class UserDefinedFontException(Exception):  # noqa: N818
    """An exception related to user font creation."""


class DictLikeObject(Protocol):
    def get(self, char: str) -> ImageData | None:
        pass


class UserDefinedMappingFont(UserDefinedFontBase):
    """The class allows the creation of user defined fonts from a set of mappings.

    .. versionadded:: 2.0.15
    """
    _glyph_renderer: UserDefinedGlyphRenderer

    def __init__(self, name: str, default_char: str, size: int, mappings: DictLikeObject,
            ascent: int | None = None, descent: int | None = None, weight: str = "normal", italic: bool = False,
            stretch: bool = False, dpi: int = 96, locale: str | None = None) -> None:
        """Initialize the default parameters of your font.

        Args:
            name:
                Name of the font. Must be unique to ensure it does not collide with any system fonts.
            default_char:
                If a character in a string is not found in the font, it will use this as fallback.
            size:
                Font size. Should be in pixels. This value will affect scaling if enabled.
            mappings:
                A dict or dict-like object with a ``get`` function.
                The ``get`` function must take a string character, and output :py:class:`~pyglet.image.ImageData` if
                found. It also must return ``None`` if no character is found.
            ascent:
                Maximum ascent above the baseline, in pixels. If None, the image height is used.
            descent:
                Maximum descent below the baseline, in pixels. Usually negative.
            weight:
                The font weight, as a string. Defaults to "normal".
            italic:
                If ``True``, this font will be used when ``italic`` is enabled for the font name.
            stretch:
                If ``True``, this font will be used when ``stretch`` is enabled for the font name.
            dpi:
                The assumed resolution of the display device, for the purposes of determining the pixel size of the
                font. Use a default of 96 for standard sizing.
            locale:
                Used to specify the locale of this font.
        """
        self.mappings = mappings
        default_image = self.mappings.get(default_char)
        if not default_image:
            msg = f"Default character '{default_char}' must exist within your mappings."
            raise UserDefinedFontException(msg)

        if ascent is None or descent is None:
            if ascent is None:
                ascent = default_image.height
            if descent is None:
                descent = 0

        super().__init__(name, default_char, size, ascent, descent, weight, italic, stretch, dpi, locale)

    def enable_scaling(self, base_size: int) -> None:
        """Enables scaling the font size.

        Args:
            base_size:
                The base size is used to calculate the ratio between new sizes and the original.
        """
        super().enable_scaling(base_size)
        glyphs, offsets = self.get_glyphs(self.default_char)
        self.ascent = glyphs[0].height
        self.descent = 0

    def get_glyphs(self, text: str) -> tuple[list[Glyph], list[GlyphPosition]]:
        """Create and return a list of Glyphs for `text`.

        If any characters do not have a known glyph representation in this font, a substitution will be made with
        the ``default_char``.
        """
        self._initialize_renderer()
        offsets = []
        glyphs = []  # glyphs that are committed.
        for c in base.get_grapheme_clusters(text):
            # Get the glyph for 'c'.  Hide tabs (Windows and Linux render
            # boxes)
            if c == "\t":
                c = " "
            if c not in self.glyphs:
                image_data = self.mappings.get(c)
                if not image_data:
                    c = self.default_char
                else:
                    self.glyphs[c] = self._glyph_renderer.render(image_data)
            glyphs.append(self.glyphs[c])
            offsets.append(base.GlyphPosition(0, 0, 0, 0))
        return glyphs, offsets


def get_scaled_user_font(font_base: UserDefinedMappingFont, size: int) -> UserDefinedMappingFont:
    """This function will return a new font instance which can scale it's size based off the original base font.

    .. note:: The scaling functionality requires the PIL library to be installed.

    .. versionadded:: 2.0.15

    Args:
        font_base:
            The base font object to create a new size from.
        size:
            The new font size. This will be scaled based on the ratio between the base size and the new size.
    """
    new_font = UserDefinedMappingFont(font_base.name, font_base.default_char, size, font_base.mappings,
                                      font_base.ascent, font_base.descent, font_base.weight, font_base.italic,
                                      font_base.stretch, font_base.dpi, font_base.locale)

    new_font.enable_scaling(font_base.size)
    return new_font


__all__ = (
    "UserDefinedFontBase",
    "UserDefinedFontException",
    "UserDefinedMappingFont",
    "get_scaled_user_font",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from pyglet.gl import GL_TRIANGLE_STRIP, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA
from pyglet.sprite import Sprite

if TYPE_CHECKING:
    from pyglet.image import AbstractImage, Animation
    from pyglet.graphics import Batch, Group
    from pyglet.text.layout import TextLayout


class NinePatch(Sprite):
    """Pseudo Nine-patch object for variable sized dialog windows.

    This class takes in a single image, splits it into 9 equal parts,
    and allows creating variable sized dialog windows without distoring
    the aspect of the edges. This is not a "real" nine-patch, as it
    does not look for any embedded markers in the image data. Instead,
    it simply splits the source image into 9 equally sized segments,
    and stretches the center segments to allow dynamic sizing. Make
    sure your input images do not have any transparent borders, as
    this can affect the final size.

    NinePatch is a subclass of :py:class:`~pyglet.sprite.Sprite`,
    and shares most of the same API. The exception is that scaling
    properties cannot be used, and will raise an exception. Instead,
    use :py:meth:`NinePatch.width` and :py:meth:`NinePatch.height`
    properties if you want to change the size after creation.
    """

    def __init__(self,
                 img: AbstractImage | Animation,
                 x: float = 0, y: float = 0, z: float = 0,
                 width: int | None = None, height: int | None = None,
                 blend_src: int = GL_SRC_ALPHA,
                 blend_dest: int = GL_ONE_MINUS_SRC_ALPHA,
                 batch: Batch | None = None,
                 group: Group | None = None):
        """Create a NinePatch instance.

        Args:
            img:
                The Image to split into segments.
            x:
                The X coordinate of the NinePatch.
            y:
                The Y coordinate of the NinePatch.
            z:
                The Z coordinate of the NinePatch.
            width:
                The desired width of the NinePatch. This must be
                greater or equal to the provided image width.
            height:
                The desired height of the NinePatch. This must be
                greater or equal to the provided image height.
            blend_src:
                OpenGL blend source mode; for example, ``GL_SRC_ALPHA``.
            blend_dest:
                OpenGL blend destination mode; for example, ``GL_ONE_MINUS_SRC_ALPHA``.
            batch:
                Optional batch to add the NinePatch to.
            group:
                Optional parent group of the NinePatch.
        """

        self._width = max(width, img.width)
        self._height = max(height, img.height)
        super().__init__(img, x, y, z, blend_src, blend_dest, batch, group)

    @classmethod
    def create_around_layout(cls,
                             img: AbstractImage | Animation,
                             layout: TextLayout,
                             border: int = 0,
                             blend_src: int = GL_SRC_ALPHA,
                             blend_dest: int = GL_ONE_MINUS_SRC_ALPHA,
                             batch: Batch | None = None,
                             group: Group | None = None):
        """Given a Label, create a NinePatch instance sized to surround it.

        A NinePatch instance will be created that is sized to the Layout's left, bottom,
        right, and top attributes. This happens at the time of creation, and is not dynamic.

        The NinePatch's ``z`` position will be set to the Layout's ``z`` position - 1. This
        will help ensure the NinePatch renders below the label *if* OpenGL depth testing is
        enabled. If not, you should provide a Group with a proper ordering to ensure the
        correct rendering order.

        Args:
            img:
                The Image to split into segments.
            layout:
                A pyglet Label or Layout instance to query the size from.
            border:
                Additional padding, in pixels, to place around the label.
            blend_src:
                OpenGL blend source mode; for example, ``GL_SRC_ALPHA``.
            blend_dest:
                OpenGL blend destination mode; for example, ``GL_ONE_MINUS_SRC_ALPHA``.
            batch:
                Optional batch to add the NinePatch to.
            group:
                Optional parent group of the NinePatch.
        """

        x = layout.left - border
        y = layout.bottom - border
        z = layout.z - 1
        width = int(layout.right - x + border)
        height = int(layout.top - y + border)
        return cls(img, x, y, z, width, height, blend_src, blend_dest, batch, group)

    def _create_vertex_list(self) -> None:
        # Vertex layout for 9 quads:
        #
        #   12----13----14----15
        #   |  /  |  /  |  /  |
        #   8-----9-----10----11
        #   |  /  |  /  |  /  |
        #   4-----5-----6-----7
        #   |  /  |  /  |  /  |
        #   0-----1-----2-----3

        # Triangle strip indices, including degenerates (duplicates)
        indices = (0, 0, 4, 1, 5, 2, 6, 3, 7,       # bottom row -->
                   11, 6, 10, 5, 9, 4, 8,           # center row <--
                   12, 9, 13, 10, 14, 11, 15, 15)   # upper row  -->

        # Get the 1/3 size of texture width & height:
        uv_x, uv_y, uv_w, uv_h = self._texture.uv
        seg_w = (uv_w - uv_x) / 3
        seg_h = (uv_h - uv_y) / 3

        # Create new UV coordinates for each of the 9 quads:
        uvs = [i for v in range(4) for h in range(4) for i in (uv_x + seg_w * h, uv_y + seg_h * v, 0)]

        self._vertex_list = self.program.vertex_list_indexed(
            16, GL_TRIANGLE_STRIP, indices, self._batch, self._group,
            position=('f', self._get_vertices()),
            colors=('Bn', self._rgba * 16),
            translate=('f', (self._x, self._y, self._z) * 16),
            scale=('f', (self._scale*self._scale_x, self._scale*self._scale_y) * 16),
            rotation=('f', (self._rotation,) * 16),
            tex_coords=('f', uvs))

    def _get_vertices(self) -> tuple:
        if not self._visible:
            return (0, 0, 0) * 16
        else:
            img = self._texture
            edge_width = img.width / 3
            edge_height = img.height / 3
            center_width = self._width - edge_width - edge_width
            center_height = self._height - edge_height - edge_height

            x0 = -img.anchor_x
            x1 = x0 + edge_width
            x2 = x1 + center_width
            x3 = x2 + edge_width
            y0 = -img.anchor_y
            y1 = y0 + edge_height
            y2 = y1 + center_height
            y3 = y2 + edge_height
            z = 0   # handled by translate attribute

            return (x0, y0, z, x1, y0, z, x2, y0, z, x3, y0, z,
                    x0, y1, z, x1, y1, z, x2, y1, z, x3, y1, z,
                    x0, y2, z, x1, y2, z, x2, y2, z, x3, y2, z,
                    x0, y3, z, x1, y3, z, x2, y3, z, x3, y3, z)

    @property
    def position(self) -> tuple[float, float, float]:
        """The (x, y, z) coordinates of the NinePatch, as a tuple."""
        return self._x, self._y, self._z

    @position.setter
    def position(self, position: tuple[float, float, float]) -> None:
        self._x, self._y, self._z = position
        self._vertex_list.translate[:] = position * 16

    @property
    def x(self) -> float:
        """X coordinate of the NinePatch."""
        return self._x

    @x.setter
    def x(self, x):
        self._x = x
        self._vertex_list.translate[:] = (x, self._y, self._z) * 16

    @property
    def y(self) -> float:
        """Y coordinate of the NinePatch."""
        return self._y

    @y.setter
    def y(self, y):
        self._y = y
        self._vertex_list.translate[:] = (self._x, y, self._z) * 16

    @property
    def z(self) -> float:
        """Z coordinate of the NinePatch."""
        return self._z

    @z.setter
    def z(self, z):
        self._z = z
        self._vertex_list.translate[:] = (self._x, self._y, z) * 16

    @property
    def rotation(self) -> float:
        """Clockwise rotation of the NinePatch, in degrees.

        The NinePatch will be rotated about its image's (anchor_x, anchor_y)
        position.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, rotation: float):
        self._rotation = rotation
        self._vertex_list.rotation[:] = (self._rotation,) * 16

    @property
    def scale(self) -> float:
        raise NotImplementedError("Not supported. Use `width`/`height` instead.")

    @property
    def scale_x(self) -> float:
        raise NotImplementedError("Not supported. Use `width` instead.")

    @property
    def scale_y(self):
        raise NotImplementedError("Not supported. Use `height` instead.")

    def update(self, *args, **kwargs):
        raise NotImplementedError("Not supported on NinePatch objects.")

    @property
    def width(self) -> float:
        """The width of the NinePatch.

        Invariant under rotation.
        """
        return self._width

    @width.setter
    def width(self, width: float):
        self._width = width
        self._update_position()

    @property
    def height(self) -> float:
        """The height of the NinePatch.

        Invariant under rotation.
        """
        return self._height

    @height.setter
    def height(self, height: float):
        self._height = height
        self._update_position()

    @property
    def opacity(self) -> int:
        """Blend opacity.

        This property sets the alpha component of the colour of the sprite's
        vertices.  With the default blend mode (see the constructor), this
        allows the sprite to be drawn with fractional opacity, blending with the
        background.

        An opacity of 255 (the default) has no effect.  An opacity of 128 will
        make the sprite appear translucent.
        """
        return self._rgba[3]

    @opacity.setter
    def opacity(self, opacity: int):
        r, g, b, _ = self._rgba
        self._rgba = r, g, b, opacity
        self._vertex_list.colors[:] = self._rgba * 16

    @property
    def color(self) -> tuple[int, int, int, int]:
        """Blend color.

        This property sets the color of the sprite's vertices. This allows the
        sprite to be drawn with a color tint.

        The color is specified as either an RGBA tuple of integers
        '(red, green, blue, opacity)' or an RGB tuple of integers
        `(red, blue, green)`.

        If there are fewer than three components, a :py:func`ValueError`
        will be raised. Each color component must be an int in the range
        0 (dark) to 255 (saturated). If any component is not an int, a
        :py:class:`TypeError` will be raised.
        """
        return self._rgba

    @color.setter
    def color(self, rgba: tuple[int, int, int, int] | tuple[int, int, int]):
        # ValueError raised by unpacking if len(rgba) < 3
        r, g, b, *a = rgba
        new_color = r, g, b, a[0] if a else 255

        # Only update if we actually have to
        if new_color != self._rgba:
            self._rgba = new_color
            self._vertex_list.colors[:] = new_color * 16



from pyglet import input


def test_parse_all_controller_mappings():
    # check all the mapping strings to confirm
    # that they can be properly parsed.
    for mapping in input.controller.mapping_list:
        guid = mapping.split(',')[0]
        assert input.controller.get_mapping(guid)



import pyglet

from pyglet.window import key

window = pyglet.window.Window(vsync=False)
fps = pyglet.window.FPSDisplay(window=window)
batch = pyglet.graphics.Batch()
label = pyglet.text.Label("Press a number from 1-0 to change the FPS. (number * 10)", x=10, y=50, batch=batch)

circle = pyglet.shapes.Circle(200, 200, 25, color=(200, 60, 130), batch=batch)
MOVEMENT_SPEED = 600


@window.event
def on_key_press(keycode, mod):
    rates = {key._1: 1/10, key._2: 1/20, key._3: 1/30, key._4: 1/40, key._5: 1/50,
             key._6: 1/60, key._7: 1/70, key._8: 1/80, key._9: 1/90, key._0: 1/100}

    if redraw_rate := rates.get(keycode):
        pyglet.clock.unschedule(window.draw)
        pyglet.clock.schedule_interval(window.draw, redraw_rate)
        global fps
        fps = pyglet.window.FPSDisplay(window=window)


@window.event
def on_refresh(dt):
    global MOVEMENT_SPEED

    circle.x += MOVEMENT_SPEED * dt

    if circle.x > window.width or circle.x < 0:
        MOVEMENT_SPEED = -MOVEMENT_SPEED

    window.clear()
    batch.draw()
    fps.draw()


pyglet.clock.schedule_interval(window.draw, 1/60)
pyglet.app.run(None)



"""
Usage

    summarize.py

Summarizes the session info collected with playmany and retry_crashes

A configuration must be active, see command configure.py

If a playlist was provided at session creation, then only the samples in the
playlist will be played, otherwise all files in samples_dir.
"""

import shutil
import sys

import extractors
import fs
from pyglet.media import instrumentation as ins
import mpexceptions
import report
import reports


def main():
    try:
        pathserv = fs.get_path_info_for_active_session()
    except mpexceptions.ExceptionUndefinedSamplesDir:
        print("The env var 'pyglet_mp_samples_dir' is not defined.")
        return 1
    except mpexceptions.ExceptionNoSessionIsActive:
        print("*** Error, no session active.")
        return 1

    count_bads = ins.CountBads()

    try:
        summarize(pathserv, count_bads)
    except mpexceptions.ExceptionAttemptToBrekReportsProtection:
        print("*** Error, attempt to overwrite reports when protect_reports "
              "is True.")
        return 1
    except mpexceptions.ExceptionNoDbgFilesPresent:
        print("*** Error, no dbg files present; maybe playmany should be run?")
        return 1
    return 0


def summarize(pathserv, count_bads):
    conf = fs.get_session_configuration(pathserv)
    if conf["dev_debug"]:
        pass
    else:
        if conf["protect_reports"]:
            raise mpexceptions.ExceptionAttemptToBreakRawDataProtection()

    session = pathserv.session
    parts = ["Summary of media_player debug info for session: %s" % session]
    filename = pathserv.special_raw_filename("samples_version")
    try:
        text = fs.txt_load(filename)
    except Exception as ex:
        print(ex)
        print("*** Error, could not read dbg/_version.txt, "
              "target dir probably not a session one.")
        sys.exit(1)
    parts.append("Samples version: %s" % text)

    overall = extractors.single_session_overall_defect_counters(pathserv, count_bads.count_bads)

    if len(overall.no_dbg_samples):
        parts.append("Samples with no .dbg recording: %d" % len(overall.no_dbg_samples))
    else:
        parts.append("All samples have a .dbg recording.")

    if len(overall.crashed_samples):
        parts.append("Samples that always crashed: %d" % len(overall.crashed_samples))
    else:
        parts.append("All samples (finally) got a report with no crash.")

    n = overall.total_relevant_samples()
    parts.append("Relevant samples (all - no_dbg - crashed): %d" % n)

    no_relevant_samples = (n == 0)
    if no_relevant_samples:
        parts.append("*** Error, no relevant samples remains to calculate"
                     " score.")
    else:
        parts.extend([
            "Naive quality number (anomalies / samples): %f" % overall.defects_per_sample(),
            "Relevant samples with perfect play: %d / %d" % (len(overall.perfect_play_samples), n),
            "Relevant samples with anomalies   : %d / %d" % (len(overall.counters_non_perfect_play_samples), n),
            ""
            ])

    # include crash info collected by retry_crashed.py
    text = reports.fragment_crash_retries(pathserv)
    parts.append(text)
    parts.append("")

    # include count of defects for samples with anomalies but no crash
    if len(overall.counters_non_perfect_play_samples):
        parts.append("Per sample defects info (always crashed and perfect plays"
                     " not listed)\n")
        for sample in sorted(overall.counters_non_perfect_play_samples.keys()):
            counters = overall.counters_non_perfect_play_samples[sample]
            text = reports.format_anomalies_counter(
                count_bads.anomalies_description, counters, sample)
            parts.append(text)
            parts.append("")

        # additionally emit reports 'anomalies' and 'all'
        for sample in overall.counters_non_perfect_play_samples.keys():
            report.report_sample(pathserv, sample, "anomalies")
            report.report_sample(pathserv, sample, "all")

    # appendix, list of samples without dbg
    if len(overall.no_dbg_samples):
        parts.append("Appendix, list of samples without dbg")
        for s in overall.no_dbg_samples:
            parts.append("\t%s" % s)
        parts.append("")

    text = "\n".join(parts)
    filename_summary = pathserv.special_report_filename("summary")
    fs.txt_save(text, filename_summary)

    # output pyglet info as captured at session creation
    src = pathserv.special_raw_filename("pyglet_info")
    dst = pathserv.special_report_filename("pyglet_info")
    shutil.copyfile(src, dst)
    src = pathserv.special_raw_filename("pyglet_hg_revision")
    dst = pathserv.special_report_filename("pyglet_hg_revision")
    try:
        shutil.copyfile(src, dst)
    except Exception:
        pass


def sysargs_to_mainargs():
    """builds main args from sys.argv"""
    if len(sys.argv) > 1 and sys.argv[1].startswith("--help"):
        print(__doc__)
        sys.exit(1)

if __name__ == '__main__':
    main()



from __future__ import annotations

import unicodedata
from typing import TYPE_CHECKING

from pyglet.libs.darwin.cocoapy import (
    CFSTR,
    ObjCClass,
    ObjCInstance,
    ObjCSubclass,
    PyObjectEncoding,
    cf,
    cfstring_to_string,
    send_super,
)
from pyglet.window import key

if TYPE_CHECKING:
    from pyglet.window.cocoa import CocoaWindow

NSArray = ObjCClass('NSArray')
NSApplication = ObjCClass('NSApplication')
NSColor = ObjCClass('NSColor')

# This custom NSTextView subclass is used for capturing all of the
# on_text, on_text_motion, and on_text_motion_select events.
class PygletTextView_Implementation:
    PygletTextView = ObjCSubclass('NSTextView', 'PygletTextView')

    @PygletTextView.method(b'@' + PyObjectEncoding)
    def initWithCocoaWindow_(self, window: CocoaWindow) -> ObjCInstance | None:
        self = ObjCInstance(send_super(self, 'init'))
        if not self:
            return None
        self._window = window
        # Interpret tab and return as raw characters
        self.setFieldEditor_(False)
        empty_string = CFSTR('')
        self.associate("empty_string", empty_string)

        # Prevent a blinking cursor in bottom left corner Python 3.9 w/ ARM mac.
        self.setInsertionPointColor_(NSColor.clearColor())
        return self

    @PygletTextView.method("v@")
    def mouseMoved_(self, event):
        # prevent cursor from being set to I-beam
        self.nextResponder().mouseMoved_(event)

    @PygletTextView.method('v')
    def dealloc(self) -> None:
        self._window = None
        cf.CFRelease(self.empty_string)
        send_super(self, 'dealloc')

    # Other functions still seem to work?
    @PygletTextView.method('v@')
    def keyDown_(self, nsevent: ObjCInstance) -> None:

        # Ignore F5 key text editing action
        # to prevent showing autocomplete suggestions
        if nsevent.keyCode() != 96:
            array = NSArray.arrayWithObject_(nsevent)
            self.interpretKeyEvents_(array)

        if not self.performKeyEquivalent_(nsevent):
            self.nextResponder().keyDown_(nsevent)

    @PygletTextView.method('v@')
    def keyUp_(self, nsevent: ObjCInstance) -> None:
        self.nextResponder().keyUp_(nsevent)

    @PygletTextView.method('v@')
    def insertText_(self, text: CFSTR) -> None:
        text = cfstring_to_string(text)
        self.setString_(self.empty_string)
        # Don't send control characters (tab, newline) as on_text events.
        if text and unicodedata.category(text[0]) != 'Cc':
            self._window.dispatch_event('on_text', text)

    @PygletTextView.method('v@')
    def insertNewline_(self, sender: ObjCInstance) -> None:
        # Distinguish between carriage return (u'\r') and enter (u'\x03').
        # Only the return key press gets sent as an on_text event.
        event = NSApplication.sharedApplication().currentEvent()
        chars = event.charactersIgnoringModifiers()
        ch = chr(chars.characterAtIndex_(0))
        if ch == '\r':
            self._window.dispatch_event('on_text', '\r')

    @PygletTextView.method('v@')
    def moveUp_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_UP)

    @PygletTextView.method('v@')
    def moveDown_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_DOWN)

    @PygletTextView.method('v@')
    def moveLeft_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_LEFT)

    @PygletTextView.method('v@')
    def moveRight_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_RIGHT)

    @PygletTextView.method('v@')
    def moveWordLeft_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_PREVIOUS_WORD)

    @PygletTextView.method('v@')
    def moveWordRight_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_NEXT_WORD)

    @PygletTextView.method('v@')
    def moveToBeginningOfLine_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_BEGINNING_OF_LINE)

    @PygletTextView.method('v@')
    def moveToEndOfLine_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_END_OF_LINE)

    @PygletTextView.method('v@')
    def scrollPageUp_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_PREVIOUS_PAGE)

    @PygletTextView.method('v@')
    def scrollPageDown_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_NEXT_PAGE)

    @PygletTextView.method('v@')
    def scrollToBeginningOfDocument_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion', key.MOTION_BEGINNING_OF_FILE)

    @PygletTextView.method('v@')
    def scrollToEndOfDocument_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion', key.MOTION_END_OF_FILE)

    @PygletTextView.method('v@')
    def deleteBackward_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_BACKSPACE)

    @PygletTextView.method('v@')
    def deleteForward_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion', key.MOTION_DELETE)

    @PygletTextView.method('v@')
    def moveUpAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_UP)

    @PygletTextView.method('v@')
    def moveDownAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_DOWN)

    @PygletTextView.method('v@')
    def moveLeftAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_LEFT)

    @PygletTextView.method('v@')
    def moveRightAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_RIGHT)

    @PygletTextView.method('v@')
    def moveWordLeftAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_PREVIOUS_WORD)

    @PygletTextView.method('v@')
    def moveWordRightAndModifySelection_(self, sender: ObjCInstance) -> None:
        self._window.dispatch_event('on_text_motion_select', key.MOTION_NEXT_WORD)

    @PygletTextView.method('v@')
    def moveToBeginningOfLineAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_BEGINNING_OF_LINE)

    @PygletTextView.method('v@')
    def moveToEndOfLineAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_END_OF_LINE)

    @PygletTextView.method('v@')
    def pageUpAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_PREVIOUS_PAGE)

    @PygletTextView.method('v@')
    def pageDownAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_NEXT_PAGE)

    @PygletTextView.method('v@')
    def moveToBeginningOfDocumentAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_BEGINNING_OF_FILE)

    @PygletTextView.method('v@')
    def moveToEndOfDocumentAndModifySelection_(self, sender: ObjCInstance) -> None:  # Mac OS X 10.6
        self._window.dispatch_event('on_text_motion_select', key.MOTION_END_OF_FILE)


PygletTextView = ObjCClass('PygletTextView')



"""In order to use the features of modern OpenGL versions, you must ask for
a specific context version. You can do this by supplying the `major_version`
and `minor_version` attributes for a GL Config.

This example creates an OpenGL 4.2 context, prints the version string to stdout,
and exits. Note that some GPU drivers may return a newer version than requested,
at their discretion.
"""
import pyglet

# Specify the OpenGL version explicitly to request 4.2 features
#
# Some other attributes relevant to OpenGL 3+:
#   forward_compatible = True       To request a context without deprecated
#                                   functionality
#   debug = True                    To request a debug context
config = pyglet.gl.Config(major_version=4, minor_version=2)

# Create a context matching the above configuration.  Will fail if
# OpenGL 4.2 is not supported by the driver.
window = pyglet.window.Window(config=config, visible=False)

# Print the version of the context created.
print('OpenGL version:', window.context.get_info().get_version())

window.close()



import pyglet

joysticks = pyglet.input.get_joysticks()
assert joysticks, 'No joystick device is connected'
joystick = joysticks[0]
joystick.open()

window = pyglet.window.Window(width=800, height=800)
batch = pyglet.graphics.Batch()

# Labels
pyglet.text.Label("Buttons:", x=15, y=window.height - 25, font_size=14, batch=batch)
pyglet.text.Label("D Pad:", x=window.width - 125, y=window.height - 25, font_size=14, batch=batch)


button_labels = []
button_shapes = []

for i in range(len(joystick.buttons)):
    rows = len(joystick.buttons) // 2
    y = window.height - 50 - 25 * (i % rows)
    x = 35 + 60 * (i // rows)
    label = pyglet.text.Label(f"{i}:", x=x, y=y, font_size=14, anchor_x='right', batch=batch)
    button_labels.append(label)
    shape = pyglet.shapes.Rectangle(x + 10, y + 1, 10, 10, color=(255, 0, 0), batch=batch)
    button_shapes.append(shape)


joystick_rect = pyglet.shapes.Rectangle(window.width // 2, window.height // 2, 10, 10, color=(255, 0, 255), batch=batch)
joystick_rect.anchor_position = joystick_rect.width // 2, joystick_rect.height // 2
d_pad_rect = pyglet.shapes.Rectangle(window.width - 75, window.height - 100, 10, 10, color=(0, 0, 255), batch=batch)


@window.event
def on_draw():
    window.clear()
    batch.draw()
    x = round((.5 * joystick.x + 1), 2) * window.width / 2
    y = round((-.5 * joystick.y + 1), 2) * window.height / 2
    rx = (.5 * joystick.rx + 1) * 60
    ry = (-.5 * joystick.ry + 1) * 60
    z = joystick.z * 50

    # Axes
    joystick_rect.position = x, y
    joystick_rect.anchor_position = joystick_rect.width // 2, joystick_rect.height // 2
    joystick_rect.width = 10 + rx + z
    joystick_rect.height = 10 + ry + z

    # Buttons
    for i in range(len(joystick.buttons)):
        rect = button_shapes[i]
        rect.color = (0, 255, 0) if joystick.buttons[i] else (255, 0, 0)

    # Hat
    d_pad_x = window.width - 100 + joystick.hat_x * 50
    d_pad_y = window.height - 100 + joystick.hat_y * 50
    d_pad_rect.position = d_pad_x, d_pad_y


pyglet.app.run()



from pyglet.media import instrumentation as ins

def test_no_unknown_state_fields_in_mp_events():
    all_fields = ins.MediaPlayerStateIterator.fields.keys()
    ok = True
    for evname in ins.mp_events:
        if evname == "version":
            continue
        for name in ins.mp_events[evname]["update_names"]:
            if name not in all_fields:
                print("Error, in evname '%s' unknown field '%s' in 'update_names'" % (evname, name))
                ok = False
        for name in ins.mp_events[evname]["other_fields"]:
            if name not in all_fields:
                print("Error, in evname '%s' unknown field '%s' in 'other_fields'" % (evname, name))
                ok = False
    if ok:
        print("test_no_unknown_state_fields_in_mp_events: passed")

def test_evname_in_mp_events_testcases():
    ok = True
    for evname in ins.mp_events:
        if evname == "version":
            continue
        for i, args in enumerate(ins.mp_events[evname]["test_cases"]):
            if evname != args[0]:
                msg = "Error, for evname %s the testase #%d does not match evname"
                print(msg % (evname, i))
                ok = False
    if ok:
        print("test_evname_in_mp_events_testcases: passed")
    

test_no_unknown_state_fields_in_mp_events()
test_evname_in_mp_events_testcases()



import pytest

from ...annotations import skip_if_continuous_integration

pytestmark = skip_if_continuous_integration()

try:
    from pyglet.media.drivers import pulse
    has_pulse = True
except ImportError:
    has_pulse = False

try:
    from pyglet.media.drivers import openal
    has_openal = True
except ImportError:
    has_openal = False

try:
    from pyglet.media.drivers import directsound
    has_directsound = True
except ImportError:
    has_directsound = False


def check_listener_defaults(listener):
    assert listener.volume == 1.0
    assert listener.position == (0, 0, 0)
    assert listener.forward_orientation == (0, 0, -1)
    assert listener.up_orientation == (0, 1, 0)


def check_modifying_values(listener):
    listener.volume = 0.5
    listener.position = (-1, 0, 0)
    listener.forward_orientation = (0, 0, 1)
    listener.up_orientation = (0, -1, 0)
    assert listener.volume == 0.5
    assert listener.position == (-1, 0, 0)
    assert listener.forward_orientation == (0, 0, 1)
    assert listener.up_orientation == (0, -1, 0)


@pytest.mark.skipif(not has_openal, reason="Test requires OpenAL")
def test_openal_listener():
    driver = openal.create_audio_driver()
    listener = driver.get_listener()
    check_listener_defaults(listener=listener)
    check_modifying_values(listener=listener)
    # Need to garbage collect the listener before the driver is deleted
    del listener


@pytest.mark.skipif(not has_pulse, reason="Test requires PulseAudio")
def test_pulse_listener():
    driver = pulse.create_audio_driver()
    listener = driver.get_listener()
    check_listener_defaults(listener=listener)
    check_modifying_values(listener=listener)
    # Need to garbage collect the listener before the driver is deleted
    del listener


@pytest.mark.skipif(not has_directsound, reason="Test requires DirectSound")
def test_directsound_listener():
    driver = directsound.create_audio_driver()
    listener = driver.get_listener()
    check_listener_defaults(listener=listener)
    check_modifying_values(listener=listener)
    # Need to garbage collect the listener before the driver is deleted
    del listener



from pytest import fixture

@fixture(autouse=True)
def monkeypatch_default_graphics_shader(monkeypatch, get_dummy_shader_program):
    """Use a dummy shader when testing non-drawing functionality"""
    monkeypatch.setattr('pyglet.graphics.get_default_shader', get_dummy_shader_program)


"""
This example demonstrates an alternative way to register an event type.
The original way looks like this:

    class Foo(EventDispatcher):
        # ...


    Foo.register_event_type('bar')

Instead of calling the classmethod register_event_type() below the implementation of your dispatcher,
you can decorate the latter with the function decorator:

    @register_event_type('bar')
    class Foo(EventDispatcher):
        # ...

All in one place.
"""

import pyglet
from pyglet.event import EventDispatcher


def register_event_type(name):
    def _reg_evt_type(cls):
        assert issubclass(cls, EventDispatcher), "Event types can only be registered on EventDispatcher subclasses"
        if not hasattr(cls, 'event_types'):
            cls.event_types = []

        cls.event_types.append(name)
        return cls

    return _reg_evt_type


def create_alarm(dt):
    alarms.dispatch_event('on_wake_up', dt)


@register_event_type('on_wake_up')
class Alarms(EventDispatcher):
    def on_wake_up(self, dt):
        print(f'Yet another {dt} seconds wasted! Wake up!')


alarms = Alarms()
pyglet.clock.schedule_interval(create_alarm, 30)
pyglet.app.run()
